class PipelineSpec(object):
    def __init__(self,
                 path=None,
                 pipeline_id=None,
                 pipeline_details=None,
                 source_details=None,
                 validation_errors=None,
                 dependencies=None,
                 cache_hash='',
                 schedule=None):
        self.path = path
        self.pipeline_id = pipeline_id
        self.pipeline_details = pipeline_details
        self.source_details = source_details
        self.validation_errors = [] if validation_errors is None else validation_errors
        self.dependencies = [] if dependencies is None else dependencies
        self.cache_hash = cache_hash
        self.schedule = schedule

    def __str__(self):
        return 'PipelineSpec({}, validation_errors={}, ' \
               'dependencies={}, cache_hash={})'\
            .format(self.pipeline_id, self.validation_errors,
                    self.dependencies, self.cache_hash)

    def __repr__(self):
        return str(self)


class BaseParser(object):

    class InvalidFileException(Exception):
        def __init__(self, short_msg, long_msg):
            self.short_msg = short_msg
            self.long_msg = long_msg

    @classmethod
    def check_filename(cls, filename):
        raise NotImplementedError()

    @classmethod
    def to_pipeline(cls, spec, fullpath):
        raise NotImplementedError()

    @staticmethod
    def replace_root_dir(path, root_dir):
        if root_dir.endswith('/'):
            root_dir = root_dir[:-1]
        if path.startswith(root_dir):
            path = '.' + path[len(root_dir):]
        return path



import logging

import redis

from datapackage_pipelines.utilities.extended_json import json


class RedisBackend(object):

    KIND = 'redis'

    def __init__(self, host=None, port=6379):
        self.redis = None
        if host is not None and len(host) > 0:
            conn = redis.StrictRedis(host=host, port=port, db=5)
            try:
                conn.ping()
                self.redis = conn
            except redis.exceptions.ConnectionError:
                logging.warning('Failed to connect to Redis, host:%s, port:%s',
                                host, port)

    def is_init(self):
        return self.redis is not None

    def get_status(self, pipeline_id):
        if self.is_init():
            status = self.redis.get(pipeline_id)
            if status is not None:
                status = json.loads(status.decode('ascii'))
                return status

    def set_status(self, pipeline_id, status):
        if self.is_init():
            self.redis.set(pipeline_id, json.dumps(status, ensure_ascii=True))

    def del_status(self, pipeline_id):
        if self.is_init():
            self.redis.delete(pipeline_id)

    def register_pipeline_id(self, pipeline_id):
        if self.is_init():
            self.redis.sadd('all-pipelines', pipeline_id.strip())

    def deregister_pipeline_id(self, pipeline_id):
        if self.is_init():
            self.redis.srem('all-pipelines', pipeline_id.strip())

    def reset(self):
        if self.is_init():
            self.redis.delete('all-pipelines')

    def all_pipeline_ids(self):
        if self.is_init():
            return [x.decode('utf-8') for x in self.redis.smembers('all-pipelines')]
        return []

    def all_statuses(self):
        if self.is_init():
            all_ids = self.redis.smembers('all-pipelines')
            pipe = self.redis.pipeline()
            for _id in sorted(all_ids):
                pipe.get(_id)
            return [json.loads(sts.decode('ascii')) for sts in pipe.execute()]
        return []



from typing import Optional

import logging

from .hook_sender import hook_sender
from .pipeline_execution import PipelineExecution

# def jcmp(a, b):
#     return json.dumps(a, sort_keys=True) != json.dumps(b, sort_keys=True)


class PipelineStatus(object):

    def __init__(self, backend,
                 pipeline_id):
        self.backend = backend
        self.pipeline_id = pipeline_id
        self.__load()

    def init(self,
             pipeline_details,
             source_spec,
             validation_errors,
             cache_hash):
        self.pipeline_details = pipeline_details
        self.source_spec = source_spec
        self.validation_errors = validation_errors
        self.cache_hash = cache_hash

    def __load(self):
        data = self.backend.get_status('PipelineStatus:' + self.pipeline_id)
        if data is None:
            data = {}
        self.pipeline_details = data.get('pipeline_details', {})
        self.source_spec = data.get('source_spec', {})
        self.validation_errors = data.get('validation_errors', [])
        self.cache_hash = data.get('cache_hash', '')
        self._execution_ids = data.get('executions', [])
        self._executions = None
        self._last_execution = None

    def __iter__(self):
        yield 'pipeline_details', self.pipeline_details,
        yield 'source_spec', self.source_spec,
        yield 'validation_errors', self.validation_errors
        yield 'cache_hash', self.cache_hash
        yield 'executions', [ex.execution_id for ex in self.executions]

    def __save(self):
        # logging.debug('SAVING PipelineStatus %s -> %r' % (self.pipeline_id, self.executions))
        self.backend.register_pipeline_id(self.pipeline_id)
        self.backend.set_status('PipelineStatus:' + self.pipeline_id, dict(self))

    def save(self):
        self.__save()

    def dirty(self):
        return self.num_executions == 0 or self.cache_hash != self.last_execution.cache_hash

    @property
    def executions(self):
        if self._executions is None:
            self._executions = [PipelineExecution.from_execution_id(self.backend, ex)
                                for ex in self._execution_ids]
        return self._executions

    @property
    def num_executions(self):
        if self._executions is not None:
            return len(self._executions)
        return len(self._execution_ids)

    @property
    def last_execution(self):
        if self._executions is not None:
            if len(self._executions) > 0:
                return self._executions[0]
        else:
            if len(self._execution_ids) > 0:
                if self._last_execution is None:
                    self._last_execution = \
                        PipelineExecution.from_execution_id(self.backend,
                                                            self._execution_ids[0])
                return self._last_execution

    def errors(self):
        if not self.runnable():
            return ['%s :%s' % tuple(err)
                    for err in self.validation_errors]
        else:
            ex = self.last_execution
            if ex is not None:
                return ex.error_log
        return []

    def runnable(self):
        return len(self.validation_errors) == 0

    def get_last_execution(self) -> Optional[PipelineExecution]:
        return self.last_execution

    def get_last_successful_execution(self) -> Optional[PipelineExecution]:
        le = self.last_execution
        if le is not None:
            if le.success:
                return le
            for ex in self.executions[1:]:
                if ex.success:
                    return ex
        return None

    def queue_execution(self, execution_id, trigger):
        last_exec = self.get_last_execution()
        if last_exec is not None and last_exec.finish_time is None:
            if last_exec.is_stale():
                last_exec.invalidate()
                last_exec.finish_execution(False, {}, ['Cancelled'])
            else:
                logging.info('%s %s is ALREADY RUNNING, BAILING', execution_id[:8], self.pipeline_id)
                return False
        # for ex in self.executions:  # type: PipelineExecution
        #     if not ex.invalidate():
        #         break
        execution = PipelineExecution(self.backend,
                                      self.pipeline_id, self.pipeline_details, self.cache_hash,
                                      trigger, execution_id, save=False)
        assert execution.queue_execution(trigger)
        self.executions.insert(0, execution)
        while len(self.executions) > 10:
            last = self.executions.pop()  # type: PipelineExecution
            last.delete()
        self.__save()
        self.update_hooks('queue', blocking=True)
        return True

    def validate_execution_id(self, execution_id):
        if len(self.executions) == 0:
            logging.info('%s NO EXISTING EXECUTIONS', execution_id[:8])
            return False
        first = self.executions[0].execution_id
        if first != execution_id:
            logging.info('%s EXECUTION ID MISMATCH (first: %s)', execution_id[:8], first)
            return False
        return True

    def start_execution(self, execution_id):
        if not self.validate_execution_id(execution_id):
            return False
        self.update_hooks('start')
        return self.executions[0].start_execution()

    def finish_execution(self, execution_id, success, stats, error_log):
        if self.validate_execution_id(execution_id):
            self.update_hooks('finish', success=success, errors=error_log, stats=stats)
            return self.executions[0].finish_execution(success, stats, error_log)
        return False

    def update_execution(self, execution_id, log, hooks=False):
        if self.validate_execution_id(execution_id):
            if hooks:
                self.update_hooks('progress', log=log)
            return self.executions[0].update_execution(log)
        return False

    def deregister(self):
        self.backend.deregister_pipeline_id(self.pipeline_id)

    def state(self):
        if not self.runnable():
            return 'INVALID'
        last_execution = self.get_last_execution()
        if last_execution is None:
            return 'INIT'
        if last_execution.success is None:
            if last_execution.start_time is None:
                return 'QUEUED'
            else:
                return 'RUNNING'
        if last_execution.success:
            return 'SUCCEEDED'
        else:
            return 'FAILED'

    def update_hooks(self, event, *, success=None, errors=None, stats=None, log=None, blocking=False):
        hooks = self.pipeline_details.get('hooks')
        if hooks is not None:
            payload = {
                'pipeline_id': self.pipeline_id,
                'event': event,
            }
            if success is not None:
                payload['success'] = success
            if errors is not None:
                payload['errors'] = errors
            if stats is not None:
                payload['stats'] = stats
            if log is not None:
                payload['log'] = log[-100:]
            for hook in hooks:
                hook_sender.send(hook, payload, blocking=blocking)



import csv
import json
import os
import isodate
import logging

from datapackage_pipelines.utilities.extended_json import DATETIME_FORMAT, DATE_FORMAT, TIME_FORMAT
from datapackage_pipelines.utilities.resources import get_path


def identity(x):
    return x


def json_dumps(x):
    return json.dumps(x, ensure_ascii=False)


class FileFormat():

    def prepare_resource(self, resource):
        for field in resource.get('schema', {}).get('fields', []):
            field.update(self.PYTHON_DIALECT.get(field['type'], {}))

    def __transform_row(self, row, fields):
        try:
            return dict((k, self.__transform_value(v, fields[k]['type']))
                        for k, v in row.items())
        except Exception:
            logging.exception('Failed to transform row %r', row)
            raise

    @classmethod
    def __transform_value(cls, value, field_type):
        if value is None:
            return cls.NULL_VALUE
        serializer = cls.SERIALIZERS.get(field_type, cls.DEFAULT_SERIALIZER)
        return serializer(value)

    def write_row(self, writer, row, fields):
        transformed_row = self.__transform_row(row, fields)
        self.write_transformed_row(writer, transformed_row, fields)


class CSVFormat(FileFormat):

    SERIALIZERS = {
        'array': json_dumps,
        'object': json_dumps,
        'datetime': lambda d: d.strftime(DATETIME_FORMAT),
        'date': lambda d: d.strftime(DATE_FORMAT),
        'time': lambda d: d.strftime(TIME_FORMAT),
        'duration': lambda d: isodate.duration_isoformat(d),
        'geopoint': lambda d: '{}, {}'.format(*d),
        'geojson': json.dumps,
        'year': lambda d: '{:04d}'.format(d),
        'yearmonth': lambda d: '{:04d}-{:02d}'.format(*d),
    }
    DEFAULT_SERIALIZER = str
    NULL_VALUE = ''

    PYTHON_DIALECT = {
        'number': {
            'decimalChar': '.',
            'groupChar': ''
        },
        'date': {
            'format': DATE_FORMAT
        },
        'time': {
            'format': TIME_FORMAT
        },
        'datetime': {
            'format': DATETIME_FORMAT
        },
    }

    def prepare_resource(self, resource):
        resource['encoding'] = 'utf-8'
        basename, _ = os.path.splitext(get_path(resource))
        resource['path'] = basename + '.csv'
        resource['format'] = 'csv'
        resource['dialect'] = dict(
            lineTerminator='\r\n',
            delimiter=',',
            doubleQuote=True,
            quoteChar='"',
            skipInitialSpace=False
        )
        super(CSVFormat, self).prepare_resource(resource)

    def initialize_file(self, file, headers):
        csv_writer = csv.DictWriter(file, headers)
        csv_writer.writeheader()
        return csv_writer

    def write_transformed_row(self, writer, transformed_row, fields):
        writer.writerow(transformed_row)

    def finalize_file(self, writer):
        pass


class JSONFormat(FileFormat):

    SERIALIZERS = {
        'datetime': lambda d: d.strftime(DATETIME_FORMAT),
        'date': lambda d: d.strftime(DATE_FORMAT),
        'time': lambda d: d.strftime(TIME_FORMAT),
        'number': float,
        'duration': lambda d: isodate.duration_isoformat(d),
        'geopoint': lambda d: list(map(float, d)),
        'yearmonth': lambda d: '{:04d}-{:02d}'.format(*d),
    }
    DEFAULT_SERIALIZER = identity
    NULL_VALUE = None

    PYTHON_DIALECT = {
        'date': {
            'format': DATE_FORMAT
        },
        'time': {
            'format': TIME_FORMAT
        },
        'datetime': {
            'format': DATETIME_FORMAT
        },
    }

    def prepare_resource(self, resource):
        resource['encoding'] = 'utf-8'
        basename, _ = os.path.splitext(get_path(resource))
        resource['path'] = basename + '.json'
        resource['format'] = 'json'
        super(JSONFormat, self).prepare_resource(resource)

    def initialize_file(self, file, headers):
        writer = file
        writer.write('[')
        writer.__first = True
        return writer

    def write_transformed_row(self, writer, transformed_row, fields):
        if not writer.__first:
            writer.write(',')
        else:
            writer.__first = False
        writer.write(json.dumps(transformed_row, sort_keys=True, ensure_ascii=True))

    def finalize_file(self, writer):
        writer.write(']')



"""
This module contains code that can be used to run automated tests of
datapackage pipelines processors

It is used internally to test the standard library processors but supports
external use as well.
"""
import os
import subprocess
import sys
import json
import mock
import importlib


class ProcessorFixtureTestsBase(object):

    def __init__(self, fixtures_path):
        self._fixtures_path = fixtures_path

    def get_tests(self):
        for dirpath, _, filenames in os.walk(self._fixtures_path):
            for filename in filenames:
                data_in, data_out, dp_out, params, processor = self._load_fixture(dirpath, filename)
                yield filename, self._get_test_func(processor, params, data_in, dp_out, data_out, filename)

    def _get_test_func(self, processor, params, data_in, dp_out, data_out, filename):
        def inner():
            return self._test_single_fixture(processor, params, data_in, dp_out, data_out,
                                             self._get_procesor_env(filename))
        return inner

    def _get_procesor_env(self, filename):
        """
        Set the environment variables for the sub-process that runs the
        processor extending classes will most likely want to set the PYTHONPATH
        variable to correct value
        you can use the filename param to setup DB or other dependencies differently for each fixture
        """
        return {}

    def _get_processor_file(self, processor):
        """
        Must be implemented by extending classes to return the path to the
        processor python file.

        :param processor: the value of the 1st part of the fixture (the name of
            the processor to test)
        :return: full path to the processor python file
        """
        raise NotImplementedError()

    def _load_fixture(self, dirpath, filename):
        with open(os.path.join(dirpath, filename), encoding='utf8') as f:
            parts = f.read().split('--\n')
        processor, params, dp_in, data_in, dp_out, data_out = parts
        processor_file = self._get_processor_file(processor)
        params = rejsonize(params)
        # Hacky way to form cwd for resource path and ignore  other formating like %Y-%m-%d
        try:
            dp_out = dp_out % ({"base": os.getcwd()})
        except: #noqa
            pass
        dp_out = rejsonize(dp_out)
        dp_in = rejsonize(dp_in)
        data_in = reline(data_in)
        data_out = reline(data_out)
        data_in = (dp_in + '\n\n' + data_in).encode('utf8')
        return data_in, data_out, dp_out, params, processor_file

    def _run_processor(self, processor, parameters, data_in, env):
        '''Run the passed `processor` and return the output'''
        process = subprocess.run([sys.executable, processor, '1',
                                  parameters, 'False', ''],
                                 input=data_in,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 env=env)
        print('\nProcessor output:\n')
        for line in process.stderr.decode('utf8').split('\n'):
            print(f'OUT> {line}')
        if process.returncode != 0:
            raise Exception(f'processor execution failed with {process.returncode}')
        return process.stdout.decode('utf8')

    def _test_single_fixture(self, processor, parameters, data_in,
                             dp_out, data_out, env):
        """Test a single processor with the given fixture parameters.
           Handle dependencies info adding and removal
        """
        output = self._run_processor(processor, parameters, b'{}\n' + data_in, env)
        self.test_fixture(output[3:], dp_out, data_out)

    @staticmethod
    def test_fixture(processor_output, dp_out, data_out):
        '''Receives processor output and performs standard tests. Can be
        overridden in subclasses.'''
        (actual_dp, *actual_data) = processor_output.split('\n\n', 1)
        if actual_dp.strip() != dp_out.strip():
            print("unexpected value for output datapackage:\n{!r}\n{!r}".format(actual_dp, dp_out),
                  file=sys.stderr)
        actual_dp = json.loads(actual_dp)
        dp_out = json.loads(dp_out)
        for ares, eres in zip(actual_dp.get('resources', []), dp_out.get('resources', [])):
            assert ares.get('schema', {}).get('fields') == eres.get('schema', {}).get('fields')
            assert ares.get('schema', {}) == eres.get('schema', {})
            assert ares == eres, 'error comparing actual:\n%r\nto expected:\n%r\n...' % (ares, eres)
        assert actual_dp == dp_out

        if len(actual_data) > 0:
            actual_data = actual_data[0]
            actual_data = actual_data.split('\n')
            expected_data = data_out.split('\n')
            assert len(actual_data) == len(expected_data), \
                "unexpected number of output lines: {}, actual_data = {}" \
                .format(len(actual_data), actual_data)
            for line_num, (actual, expected) in enumerate(zip(actual_data,
                                                              expected_data)):
                line_msg = "output line {}".format(line_num)
                if len(actual) == 0:
                    assert len(expected) == 0, \
                        "{}: did not get any data (but expected some)" \
                        .format(line_msg)
                else:
                    rj_actual = rejsonize(actual)
                    if rj_actual != rejsonize(expected):
                        print("{}: unexpected data: {} (expected {})".format(line_msg, rj_actual, expected),
                              file=sys.stderr)
                    assert json.loads(actual) == json.loads(expected), \
                        "a: %r, e: %r" % (json.loads(actual), json.loads(expected))


def rejsonize(s):
    return json.dumps(json.loads(s), sort_keys=True, ensure_ascii=True)


def reline(data):
    data = data.strip().split('\n')
    out = ''
    buf = ''
    for line in data:
        if not line.strip():
            out += '\n'
        buf += line
        try:
            buf = json.loads(buf)
        except Exception:
            continue
        out += json.dumps(buf, sort_keys=True, ensure_ascii=True) + '\n'
        buf = ''
    return out


@mock.patch('datapackage_pipelines.wrapper.ingest')
@mock.patch('datapackage_pipelines.wrapper.spew')
def mock_processor_test(processor, ingest_tuple, mock_spew, mock_ingest):
    '''Helper function returns the `spew` for a given processor with a given
    `ingest` tuple.'''

    # Mock all calls to `ingest` to return `ingest_tuple`
    mock_ingest.return_value = ingest_tuple

    # Call processor
    file_path = processor
    module_name, _ = os.path.splitext(os.path.basename(file_path))
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Our processor called `spew`. Return the args it was called with.
    return mock_spew.call_args



from dataflows import Flow, duplicate
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow, load_lazy_json


def flow(parameters):
    return Flow(
        load_lazy_json(parameters.get('source')),
        duplicate(
            parameters.get('source'),
            parameters.get('target-name'),
            parameters.get('target-path'),
            parameters.get('batch_size', 1000)
        )
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



import warnings
import zipfile

from datapackage_pipelines.lib.dump.dumper_base import FileDumper


class ZipDumper(FileDumper):

    def initialize(self, params):
        super(ZipDumper, self).initialize(params)
        out_filename = open(params['out-file'], 'wb')
        self.zip_file = zipfile.ZipFile(out_filename, 'w')

    def write_file_to_output(self, filename, path):
        self.zip_file.write(filename, arcname=path,
                            compress_type=zipfile.ZIP_DEFLATED)

    def finalize(self):
        self.zip_file.close()
        super(ZipDumper, self).finalize()


if __name__ == '__main__':
    warnings.warn(
        'dump.to_zip will be removed in the future, use "dump_to_zip" instead',
        DeprecationWarning
    )
    ZipDumper()()



STATS_DPP_KEY = '.dpp'
STATS_OUT_DP_URL_KEY = 'out-datapackage-url'


def user_facing_stats(stats):
    if stats is not None and isinstance(stats, dict):
        return dict((k, v) for k, v in stats.items() if k != STATS_DPP_KEY)
    return None



import os

from celery import Celery


REGULAR_TASK_NAME = 'datapackage_pipelines.celery_tasks.celery_tasks' + \
                        '.execute_pipeline_task'
SCHEDULED_TASK_NAME = 'datapackage_pipelines.celery_tasks.celery_tasks' + \
                        '.execute_scheduled_pipeline'
MANAGEMENT_TASK_NAME = 'datapackage_pipelines.celery_tasks.celery_tasks' + \
                        '.update_pipelines'


def get_celery_app(**kwargs):
    celery_app = Celery('dpp')

    broker = os.environ.get('DPP_CELERY_BROKER', 'redis://localhost:6379/6')

    conf = dict(
        CELERY_TIMEZONE='UTC',
        CELERY_REDIRECT_STDOUTS=False,
        BROKER_URL=broker,
        CELERY_RESULT_BACKEND=broker,
        CELERYD_LOG_LEVEL="DEBUG",
        CELERY_TASK_SERIALIZER='json',
        CELERY_RESULT_SERIALIZER='json',
        CELERY_ACCEPT_CONTENT=['json'],
        CELERYD_LOG_FORMAT='[%(asctime)s: %(levelname)s/%(processName)s(%(process)d)] %(message)s',
        CELERY_ROUTES={
            REGULAR_TASK_NAME: {'queue': 'datapackage-pipelines'},
            SCHEDULED_TASK_NAME: {'queue': 'datapackage-pipelines-management'},
            MANAGEMENT_TASK_NAME: {'queue': 'datapackage-pipelines-management'},
        }
    )
    conf.update(kwargs)

    celery_app.conf.update(**conf)

    return celery_app



from tabulator.parser import Parser
from tabulator.helpers import reset_stream


class TXTParser(Parser):
    """Parser to parse TXT data format.
    """

    # Public

    options = []

    def __init__(self, loader, **options):
        super(TXTParser, self).__init__(loader, **options)

        # Set attributes
        self.__options = options
        self.__extended_rows = None
        self.__loader = loader
        self.__chars = None
        self.__encoding = None

    @property
    def closed(self):
        return self.__chars is None or self.__chars.closed

    def open(self, source, encoding=None, force_parse=False):
        self.close()
        self.__chars = self.__loader.load(source, encoding)
        self.__encoding = getattr(self.__chars, 'encoding', encoding)
        if self.__encoding:
            self.__encoding.lower()
        self.reset()

    def close(self):
        if not self.closed:
            self.__chars.close()

    def reset(self):
        reset_stream(self.__chars)
        self.__extended_rows = self.__iter_extended_rows()

    @property
    def extended_rows(self):
        return self.__extended_rows

    @property
    def encoding(self):
        return self.__encoding

    # Private

    def __iter_extended_rows(self):
        for number, line in enumerate(self.__chars, start=1):
            if line.endswith('\n'):
                line = line[:-1]
            yield (number, None, [line])



from dataflows import Flow, dump_to_zip
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow


def flow(parameters: dict):
    out_file = parameters.pop('out-file')
    return Flow(
        dump_to_zip(
            out_file,
            **parameters
        )
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



import os
from typing import Iterator

from .base_parser import BaseParser, PipelineSpec


class BasicPipelineParser(BaseParser):

    SPEC_FILENAME = 'pipeline-spec.yaml'

    @classmethod
    def check_filename(cls, filename):
        return filename == cls.SPEC_FILENAME

    @classmethod
    def to_pipeline(cls, spec, fullpath, root_dir='.') -> Iterator[PipelineSpec]:
        dirpath = os.path.dirname(fullpath)

        for pipeline_id, pipeline_details in spec.items():
            pipeline_id = os.path.join(dirpath, pipeline_id)
            pipeline_id = cls.replace_root_dir(pipeline_id, root_dir)
            yield PipelineSpec(path=dirpath,
                               pipeline_id=pipeline_id,
                               pipeline_details=pipeline_details)



from dataflows import Flow, set_type, validate, delete_fields
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow


def flow(parameters):
    resources = parameters.get('resources')
    regex = parameters.get('regex', True)
    if 'types' in parameters:
        return Flow(
            *[
                set_type(name, resources=resources, regex=regex, **options)
                if options is not None else
                delete_fields([name], resources=resources)
                for name, options in parameters['types'].items()
            ]
        )
    else:
        return Flow(
            validate()
        )


if __name__ == '__main__':
    with ingest() as ctx:
        print(flow(ctx.parameters).chain)
        spew_flow(flow(ctx.parameters), ctx)



# -*- coding: utf-8 -*-
import io
import os

from .specs import pipelines
from .manager import execute_pipeline

VERSION_FILE = os.path.join(os.path.dirname(__file__), 'VERSION')

__version__ = io.open(VERSION_FILE, encoding='utf-8').readline().strip()



from dataflows import Flow, update_resource
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow


def flow(parameters):
    resources = parameters.get('resources', None)
    metadata = parameters.pop('metadata', {})
    return Flow(
        update_resource(resources, **metadata),
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



# pylama:skip=1
from typing import NamedTuple


class SpecError(NamedTuple):
    short_msg: str
    long_msg: str

    def __str__(self):
        return '{}: {}'.format(self.short_msg, self.long_msg)



from dataflows import Flow, load, update_package
from dataflows.helpers.resource_matcher import ResourceMatcher

from datapackage_pipelines.wrapper import ProcessorContext
from datapackage_pipelines.utilities.extended_json import LazyJsonLine


def load_lazy_json(resources):

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        yield package.pkg
        for rows in package:
            if matcher.match(rows.res.name):
                yield (
                    row.inner
                    if isinstance(row, LazyJsonLine)
                    else row
                    for row in rows
                )
            else:
                yield rows

    return func


class MergeableStats():
    def __init__(self, ds_stats, ctx_stats):
        self.ds_stats = ds_stats
        self.ctx_stats = ctx_stats

    def __iter__(self):
        if self.ds_stats is not None:
            for x in self.ds_stats:
                yield from x.items()
        if self.ctx_stats is not None:
            yield from self.ctx_stats.items()


def spew_flow(flow, ctx: ProcessorContext):
    flow = Flow(
        update_package(**ctx.datapackage),
        load((ctx.datapackage, ctx.resource_iterator)),
        flow,
    )
    datastream = flow.datastream()
    ctx.datapackage = datastream.dp.descriptor
    ctx.resource_iterator = datastream.res_iter
    ctx.stats = MergeableStats(datastream.stats, ctx.stats)



import os

from ..resolver import resolve_generator
from ..errors import SpecError
from .base_parser import BaseParser, PipelineSpec


class SourceSpecPipelineParser(BaseParser):

    SOURCE_FILENAME_SUFFIX = '.source-spec.yaml'

    @classmethod
    def check_filename(cls, filename):
        return filename.endswith(cls.SOURCE_FILENAME_SUFFIX)

    @classmethod
    def fix_dependency(cls, dep, dirpath, root_dir):
        if dep.startswith('./'):
            dep = dep[2:]
        return os.path.join(cls.replace_root_dir(dirpath, root_dir), dep)

    @classmethod
    def to_pipeline(cls, source_spec, fullpath, root_dir='.'):
        filename = os.path.basename(fullpath)
        dirpath = os.path.dirname(fullpath)

        module_name = filename[:-len(cls.SOURCE_FILENAME_SUFFIX)]
        pipeline_id = os.path.join(dirpath, module_name)
        generator = resolve_generator(module_name)

        if generator is None:
            message = 'Unknown source description kind "{}" in {}' \
                .format(module_name, fullpath)
            error = SpecError('Unknown source kind', message)
            yield PipelineSpec(pipeline_id=module_name,
                               path=dirpath,
                               validation_errors=[error],
                               pipeline_details={'pipeline': []})
            return

        base = cls.replace_root_dir(dirpath, root_dir)
        if generator.internal_validate(source_spec):
            try:
                spec = generator.internal_generate(source_spec, base)
                for pipeline_id, pipeline_details in spec:
                    if pipeline_id[0] == ':' and pipeline_id[-1] == ':':
                        module = pipeline_id[1:-1]
                        filename = module + cls.SOURCE_FILENAME_SUFFIX
                        yield from cls.to_pipeline(pipeline_details,
                                                   os.path.join(dirpath, filename))
                    else:
                        yield PipelineSpec(path=pipeline_details.get('__path', dirpath),
                                           pipeline_id=pipeline_id,
                                           pipeline_details=pipeline_details,
                                           source_details=source_spec)
            except Exception as e:
                message = '"{}" in {}' \
                    .format(e, fullpath)
                error = SpecError('Error converting source', message)
                yield PipelineSpec(pipeline_id=pipeline_id,
                                   path=dirpath, validation_errors=[error],
                                   pipeline_details={'pipeline': []})
        else:
            message = 'Invalid source description for "{}" in {}' \
                .format(module_name, fullpath)
            error = SpecError('Invalid Source', message)
            yield PipelineSpec(pipeline_id=pipeline_id,
                               path=dirpath,
                               validation_errors=[error],
                               pipeline_details={'pipeline': []})



import logging
import os

from ..utilities.execution_id import gen_execution_id

from ..status import status_mgr
from ..status.pipeline_status import PipelineStatus

from .dependency_manager import DependencyManager
from .celery_common import get_celery_app
from ..specs import pipelines, PipelineSpec
from ..manager.tasks import execute_pipeline


celery_app = get_celery_app()


executed_hashes = {}
# dependencies = {}
# dependents = {}
already_init = False
dep_mgr = None


def get_dep_mgr() -> DependencyManager:
    global dep_mgr
    if dep_mgr is None:
        dep_mgr = DependencyManager()
    return dep_mgr


def build_dependents():
    dm = get_dep_mgr()
    for spec in pipelines():  # type: PipelineSpec
        dm.update(spec)


def collect_dependencies(pipeline_ids):
    if pipeline_ids is None:
        return None
    dm = get_dep_mgr()
    ret = set()
    for pipeline_id in pipeline_ids:
        ret.add(pipeline_id)
        deps = dm.get_dependencies(pipeline_id)
        if deps is not None:
            ret.update(collect_dependencies(deps))
    return ret


def queue_pipeline(ps: PipelineStatus, spec: PipelineSpec, trigger):
    eid = gen_execution_id()
    logging.info('%s QUEUEING %s task %s', eid[:8], trigger.upper(), spec.pipeline_id)
    if ps.queue_execution(eid, trigger):
        execute_pipeline_task.delay(spec.pipeline_id,
                                    spec.pipeline_details,
                                    spec.path,
                                    trigger,
                                    eid)
        return True
    return False


def execute_update_pipelines():
    update_pipelines.delay('update', None, None)


@celery_app.task
def update_pipelines(action, completed_pipeline_id, completed_trigger):
    # action=scheduled: start a scheduled pipeline
    # action=update: iterate over all pipelines, register new ones, trigger dirty ones
    # action=complete: iterate over all pipelines, trigger dependencies
    # completed_pipeline_id: pipeline id that had just completed (or scheduled pipeline)
    # completed_trigger: the trigger for the pipeline that had just completed (when applicable)

    logging.info("Update Pipelines (%s)", action)
    status_all_pipeline_ids = set(status_mgr().all_pipeline_ids())
    executed_count = 0
    all_pipeline_ids = set()
    dm = get_dep_mgr()

    if action == 'complete':
        filter = collect_dependencies(dm.get_dependents(completed_pipeline_id))
        logging.info("DEPENDENTS Pipeline: %s <- %s", completed_pipeline_id, filter)
    else:
        filter = ('',)

    for spec in pipelines(filter):  # type: PipelineSpec
        all_pipeline_ids.add(spec.pipeline_id)
        ps = status_mgr().get(spec.pipeline_id)
        ps.init(spec.pipeline_details,
                spec.source_details,
                spec.validation_errors,
                spec.cache_hash)

        if action == 'init':
            psle = ps.get_last_execution()
            if psle is not None and not psle.finish_time:
                psle.invalidate()
                psle.finish_execution(False, {}, ['Cancelled'])

        elif action == 'update':
            if spec.pipeline_id not in status_all_pipeline_ids:
                dm.update(spec)
                logging.info("NEW Pipeline: %s", spec)
                ps.save()
            logging.debug('Pipeline: %s (dirty: %s, #ex=%s, ch=%s ex0-cache=%s)',
                          spec.pipeline_id, ps.dirty(), len(ps.executions), ps.cache_hash,
                          ps.executions[0].cache_hash if len(ps.executions) > 0 else None)

        elif action == 'complete':
            if completed_pipeline_id in spec.dependencies:
                logging.info("DEPENDENT Pipeline: %s (%d errors) (from ...%s), trigger=%s",
                             spec.pipeline_id, len(spec.validation_errors),
                             os.path.basename(completed_pipeline_id),
                             completed_trigger)
                ps.save()
            else:
                continue

        elif action == 'scheduled':
            if completed_pipeline_id != spec.pipeline_id:
                continue

        last_successful = ps.state() == 'SUCCEEDED'
        if ps.runnable() and \
                (ps.dirty() or
                 completed_trigger or
                 (action == 'init' and not last_successful)):
            queued = queue_pipeline(ps, spec,
                                    'dirty-task-%s' % action
                                    if completed_trigger is None
                                    else completed_trigger)
            if queued:
                executed_count += 1
                if executed_count == 4 and action == 'update':
                    # Limit ops on update only
                    break

    if executed_count == 0 and action == 'update':
        extra_pipelines = status_all_pipeline_ids.difference(all_pipeline_ids)
        for pipeline_id in extra_pipelines:
            logging.info("Removing Pipeline: %s", pipeline_id)
            status_mgr().deregister(pipeline_id)
            dm.remove(pipeline_id)


@celery_app.task
def execute_scheduled_pipeline(pipeline_id):
    for spec in pipelines([pipeline_id]):
        if spec.pipeline_id == pipeline_id:
            logging.info('Running scheduled pipeline %s, with schedule %s (%r)',
                         pipeline_id, spec.schedule,
                         spec.pipeline_details.get('schedule'))
            update_pipelines.delay('scheduled', pipeline_id, 'scheduled')


@celery_app.task
def execute_pipeline_task(pipeline_id,
                          pipeline_details,
                          pipeline_cwd,
                          trigger,
                          execution_id):

    spec = PipelineSpec(pipeline_id=pipeline_id,
                        pipeline_details=pipeline_details,
                        path=pipeline_cwd)
    success, _, _ = \
        execute_pipeline(spec,
                         execution_id,
                         trigger,
                         False)

    if success:
        has_dependents = len(get_dep_mgr().get_dependents(spec.pipeline_id)) > 0
        logging.info('HAS DEPENDENTS %s: %s', pipeline_id, has_dependents)
        if has_dependents:
            update_pipelines.delay('complete', pipeline_id, trigger)



import os

from .backend_redis import RedisBackend
from .backend_filesystem import FilesystemBackend
from .pipeline_status import PipelineStatus


class StatusManager(object):

    def __init__(self, *, host=None, port=6379, root_dir='.'):
        self._host = host
        self._port = port
        self._backend = None
        self._root_dir = root_dir

    @property
    def backend(self):
        if self._backend is None:
            redis = RedisBackend(self._host, self._port)
            self._backend = redis if redis.is_init() else FilesystemBackend(self._root_dir)
        return self._backend

    def get_errors(self, _id):
        ex = self.get(_id).get_last_execution()
        if ex is not None:
            return ex.error_log
        return []

    def initialize(self):
        self.backend.reset()

    def get(self, _id) -> PipelineStatus:
        return PipelineStatus(self.backend, _id)

    def all_statuses(self):
        return self.backend.all_statuses()

    def all_pipeline_ids(self):
        return self.backend.all_pipeline_ids()

    def deregister(self, pipeline_id):
        return self.get(pipeline_id).deregister()


_status = None
_root_dir = None


def status_mgr(root_dir='.') -> StatusManager:
    global _status
    global _root_dir

    if _status is not None and _root_dir == root_dir:
        return _status
    _root_dir = root_dir
    _status = StatusManager(host=os.environ.get('DPP_REDIS_HOST'), root_dir=root_dir)
    return _status



import sys
import os
import json
import concurrent
import subprocess
import threading
from queue import Queue

from collections import namedtuple

from ..utilities.execution_id import gen_execution_id

from ..specs import pipelines, PipelineSpec #noqa
from ..status import status_mgr
from ..lib.internal.sink import SINK_MAGIC
from .tasks import execute_pipeline, finalize


ExecutionResult = namedtuple('ExecutionResult',
                             ['pipeline_id', 'success', 'stats', 'errors'])

ProgressReport = namedtuple('ProgressReport',
                            ['pipeline_id', 'row_count', 'success', 'errors', 'stats'])

MAGIC = 'INFO    :(sink): '+SINK_MAGIC


def remote_execute_pipeline(spec, root_dir, use_cache, verbose, progress_report_queue):
    args = ['dpp',
            'run',
            '--slave',
            '--use-cache' if use_cache else '--no-use-cache',
            spec.pipeline_id]
    popen = subprocess.Popen(args, encoding='utf8',
                             stderr=subprocess.PIPE,
                             stdout=subprocess.DEVNULL,
                             env=os.environ.copy(),
                             cwd=root_dir)
    progress = 0
    lines = []
    for line in popen.stderr:
        if len(line) == 0:
            continue
        if line.startswith(MAGIC):
            if progress_report_queue is not None:
                progress = int(line[len(MAGIC):].strip())
                progress_report_queue.put(ProgressReport(spec.pipeline_id, progress, None, None, None))
            continue
        while len(lines) > 0:
            log = lines.pop(0)
            if verbose:
                sys.stderr.write('[%s:%s] >>> %s' %
                                 (spec.pipeline_id, threading.current_thread().name, log))
        lines.append(line)
    if len(lines) > 0:
        results = lines.pop(0)
    else:
        if progress_report_queue is not None:
            progress_report_queue.put(ProgressReport(spec.pipeline_id,
                                                     progress,
                                                     False,
                                                     ['Empty'],
                                                     None
                                                     ))
        return (spec.pipeline_id, False, {}, ['Empty'])
    try:
        results = json.loads(results)
        if progress_report_queue is not None:
            progress_report_queue.put(ProgressReport(spec.pipeline_id,
                                                     progress,
                                                     results[0]['success'],
                                                     results[0]['errors'],
                                                     results[0]['stats']
                                                     ))
    except json.decoder.JSONDecodeError:
        if verbose:
            sys.stderr.write('[%s:%s] >>> %s' % (spec.pipeline_id, threading.current_thread().name, results))
        if progress_report_queue is not None:
            progress_report_queue.put(ProgressReport(spec.pipeline_id,
                                                     progress,
                                                     False,
                                                     ['Crashed', results],
                                                     None
                                                     ))
        return (spec.pipeline_id, False, {}, [results])

    results = results[0]
    return (spec.pipeline_id,
            results['success'],
            results['stats'],
            results['errors'])


def progress_report_handler(callback, queue):
    while True:
        report = queue.get()
        if report is not None:
            callback(report)
        else:
            return


def match_pipeline_id(arg, pipeline_id):
    if arg == 'all':
        return True
    elif ',' in arg:
        return any((match_pipeline_id(a, pipeline_id) for a in arg.split(',')))
    elif arg.endswith('%'):
        return pipeline_id.startswith(arg[:-1])
    else:
        return pipeline_id == arg


def specs_to_execute(argument, root_dir, status_manager, ignore_missing_deps, dirty, results):

    pending = set()
    executed = set()
    completed = set()

    for spec in pipelines(ignore_missing_deps=ignore_missing_deps,
                          root_dir=root_dir, status_manager=status_manager):
        if match_pipeline_id(argument, spec.pipeline_id):

            # If only dirty was requested
            if dirty:
                ps = status_manager.get(spec.pipeline_id)
                if not ps.dirty():
                    continue

            pending.add(spec.pipeline_id)

    while len(pending) > 0:
        to_yield = None
        for spec in pipelines(ignore_missing_deps=ignore_missing_deps,
                              root_dir=root_dir, status_manager=status_manager):
            pipeline_id = spec.pipeline_id
            if pipeline_id not in pending:
                continue

            unresolved = set(spec.dependencies) - completed
            if len(unresolved) == 0:
                to_yield = spec
                break

            unresolved = unresolved - executed - pending
            if len(unresolved) > 0:
                # No point in waiting for dependencies that will never be resolved
                to_yield = spec
                break

        if to_yield is not None:
            executed.add(to_yield.pipeline_id)
            pending.remove(to_yield.pipeline_id)
        completed_pipeline_id = yield(to_yield)
        if completed_pipeline_id is not None:
            completed.add(completed_pipeline_id)

    yield None


def run_pipelines(pipeline_id_pattern,
                  root_dir,
                  use_cache=True,
                  dirty=False,
                  force=False,
                  concurrency=1,
                  verbose_logs=True,
                  progress_cb=None,
                  slave=False):
    """Run a pipeline by pipeline-id.
       pipeline-id supports the '%' wildcard for any-suffix matching.
       Use 'all' or '%' for running all pipelines"""
    with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency,
                                               thread_name_prefix='T') as executor:
        try:
            results = []
            pending_futures = set()
            done_futures = set()
            finished_futures = []
            progress_thread = None
            progress_queue = None
            status_manager = status_mgr(root_dir)

            if progress_cb is not None:
                progress_queue = Queue()
                progress_thread = threading.Thread(target=progress_report_handler, args=(progress_cb, progress_queue))
                progress_thread.start()

            all_specs = specs_to_execute(pipeline_id_pattern, root_dir, status_manager, force, dirty, results)

            while True:

                done = None
                if len(done_futures) > 0:
                    done = done_futures.pop()
                    finished_futures.append(done)
                    done = done.result()[0]

                try:
                    spec = all_specs.send(done)
                except StopIteration:
                    spec = None

                if spec is None:
                    # Wait for all runners to idle...
                    if len(done_futures) == 0:
                        if len(pending_futures) > 0:
                            done_futures, pending_futures = \
                                concurrent.futures.wait(pending_futures,
                                                        return_when=concurrent.futures.FIRST_COMPLETED)
                            continue
                        else:
                            break
                    else:
                        continue

                if len(spec.validation_errors) > 0:
                    results.append(
                        ExecutionResult(spec.pipeline_id,
                                        False,
                                        {},
                                        ['init'] + list(map(str, spec.validation_errors)))
                    )
                    continue

                if slave:
                    ps = status_manager.get(spec.pipeline_id)
                    ps.init(spec.pipeline_details,
                            spec.source_details,
                            spec.validation_errors,
                            spec.cache_hash)
                    eid = gen_execution_id()
                    if ps.queue_execution(eid, 'manual'):
                        success, stats, errors = \
                            execute_pipeline(spec, eid,
                                             use_cache=use_cache)
                        results.append(ExecutionResult(
                            spec.pipeline_id,
                            success,
                            stats,
                            errors
                        ))
                    else:
                        results.append(
                            ExecutionResult(spec.pipeline_id,
                                            False,
                                            None,
                                            ['Already Running'])
                        )

                else:
                    f = executor.submit(remote_execute_pipeline,
                                        spec,
                                        root_dir,
                                        use_cache,
                                        verbose_logs,
                                        progress_queue)
                    pending_futures.add(f)

            for f in finished_futures:
                ret = f.result()
                results.append(ExecutionResult(*ret))

        except KeyboardInterrupt:
            pass
        finally:
            if slave:
                finalize()

            if progress_thread is not None:
                progress_queue.put(None)
                progress_thread.join()

    return results



import time


class PipelineExecution(object):

    def __init__(self, backend, pipeline_id, pipeline_details, cache_hash, trigger, execution_id,
                 log='', queue_time=None, start_time=None, finish_time=None, success=None,
                 stats=None, error_log=None,
                 save=True):
        self.backend = backend
        self.pipeline_id = pipeline_id
        self.pipeline_details = pipeline_details
        self.cache_hash = cache_hash
        self.trigger = trigger
        self.execution_id = execution_id
        self.log = log
        self.queue_time = queue_time
        self.start_time = start_time
        self.finish_time = finish_time
        self.success = success
        self.stats = stats or {}
        self.error_log = error_log or []
        if save:
            self.__save()

    @staticmethod
    def from_execution_id(backend, execution_id):
        data = backend.get_status('PipelineExecution:' + execution_id)
        return PipelineExecution(
            backend,
            data['pipeline_id'],
            data['pipeline_details'],
            data['cache_hash'],
            data['trigger'],
            data['execution_id'],
            log=data['log'],
            queue_time=data['queue_time'],
            start_time=data['start_time'],
            finish_time=data['finish_time'],
            success=data['success'],
            stats=data['stats'],
            error_log=data['error_log'],
            save=False
        )

    def __iter__(self):
        yield 'pipeline_id', self.pipeline_id,
        yield 'pipeline_details', self.pipeline_details,
        yield 'cache_hash', self.cache_hash
        yield 'trigger', self.trigger
        yield 'execution_id', self.execution_id
        yield 'log', self.log
        yield 'queue_time', self.queue_time
        yield 'start_time', self.start_time
        yield 'finish_time', self.finish_time
        yield 'success', self.success
        yield 'stats', self.stats
        yield 'error_log', self.error_log

    def __save(self):
        # logging.debug('SAVING PipelineExecution %s/%s -> %r' % (self.pipeline_id, self.execution_id, dict(self)))
        self.backend.set_status('PipelineExecution:' + self.execution_id, dict(self))

    def queue_execution(self, trigger):
        if self.queue_time is not None:
            return False
        self.queue_time = time.time()
        self.trigger = trigger
        self.__save()
        return True

    def start_execution(self):
        if self.queue_time is None or self.start_time is not None:
            return False
        self.start_time = time.time()
        self.__save()
        return True

    def finish_execution(self, success, stats, error_log):
        if self.queue_time is None or self.start_time is None:
            return False
        self.finish_time = time.time()
        self.success = success
        self.stats = stats
        self.error_log = error_log
        self.__save()
        return True

    def update_execution(self, log):
        if self.queue_time is None or self.start_time is None or self.finish_time is not None:
            return False
        self.log = '\n'.join(log)
        self.__save()
        return True

    def invalidate(self):
        now = time.time()
        self.queue_time = self.queue_time or now
        self.start_time = self.start_time or now
        self.finish_time = self.finish_time or now
        self.__save()
        return True

    def delete(self):
        self.backend.del_status(self.execution_id)

    def is_stale(self):
        long_ago = time.time() - 86400  # a day ago
        if self.start_time is not None and self.start_time < long_ago:
            return True
        if self.queue_time is not None and self.queue_time < long_ago:
            return True



import os
import logging
import time
from datetime import date
import itertools
from decimal import Decimal
import requests

import tabulator

from tableschema import Schema

from dataflows.helpers.resource_matcher import ResourceMatcher

from datapackage_pipelines.wrapper import ingest, spew
from datapackage_pipelines.utilities.resources import streamable, PATH_PLACEHOLDER, get_path, \
    PROP_STREAMED_FROM, PROP_STREAMING, streaming
from datapackage_pipelines.utilities.extended_json import json
from datapackage_pipelines.utilities.tabulator_txt_parser import TXTParser


def _tostr(value):
    if isinstance(value, str):
        return value
    elif value is None:
        return ''
    elif isinstance(value, (int, float, bool, Decimal)):
        return str(value)
    elif isinstance(value, date):
        return value.isoformat()
    elif isinstance(value, (list, dict)):
        return json.dumps(value)

    assert False, "Internal error - don't know how to handle %r of type %r" % (value, type(value))


def _reader(opener, _url, max_row=-1):
    yield None
    filename = os.path.basename(_url)
    logging.info('%s: OPENING %s', filename, _url)
    _schema, _headers, _, _reader, _close = opener()
    num_headers = len(_headers)
    i = 0
    for i, row in enumerate(_reader):
        row = [_tostr(x).strip() for x in row]
        values = set(row)
        if len(values) == 1 and '' in values:
            # In case of empty rows, just skip them
            continue
        output = dict(zip(_headers, _schema.cast_row(row[:num_headers])))
        yield output

        i += 1
        if i % 10000 == 0:
            logging.info('%s: %d rows', filename, i)
            # break
        if i == max_row:
            break

    _close()
    logging.info('%s: TOTAL %d rows', filename, i)


def dedupe(headers):
    _dedupped_headers = []
    for hdr in headers:
        if hdr is None:
            break
        hdr = str(hdr).strip()
        if len(hdr) == 0:
            continue
        if hdr in _dedupped_headers:
            i = 0
            deduped_hdr = hdr
            while deduped_hdr in _dedupped_headers:
                i += 1
                deduped_hdr = '%s_%s' % (hdr, i)
            hdr = deduped_hdr
        _dedupped_headers.append(hdr)
    return _dedupped_headers


def add_constants(extra_headers, extra_values, columns):
    def _func(extended_rows):
        for number, headers, row in extended_rows:
            if columns is None:
                row = row[:len(headers)] + extra_values
            else:
                row = row[:columns] + extra_values
            yield number, headers + extra_headers, row
    return _func


def suffix_remover(format):
    def empty_row(row):
        return all(v is None or (isinstance(v, str) and v == '')
                   for v in row)

    def _func(extended_rows):
        suffix = False
        for number, headers, row in extended_rows:
            if format == 'txt':
                yield number, headers, row
                continue

            if suffix:
                if len(row) >= len(headers):
                    logging.warning('Expected an empty row, but got %r instead' % row)
            else:
                if empty_row(row):
                    continue
                elif len(row) < len(headers):
                    suffix = True
                else:
                    yield number, headers, row

    return _func


def stream_reader(_resource, _url, _ignore_missing, limit_rows, http_headers):
    def get_opener(__url, __resource, columns=None):
        _columns = columns

        def opener():
            _params = dict(headers=1)
            format = __resource.get("format")
            if format == "txt":
                # datapackage-pipelines processing requires having a header row
                # for txt format we add a single "data" column
                _params["headers"] = ["data"]
                _params["custom_parsers"] = {"txt": TXTParser}
                _params["allow_html"] = True
            else:
                if format is None:
                    _, format = tabulator.helpers.detect_scheme_and_format(__url)
                if format in tabulator.config.SUPPORTED_COMPRESSION:
                    format = None
                else:
                    try:
                        parser_cls = tabulator.helpers.import_attribute(tabulator.config.PARSERS[format])
                    except KeyError:
                        logging.error("Unknown format %r", format)
                        raise
                    _params.update(
                        dict(x for x in __resource.items()
                             if x[0] in parser_cls.options))
                _params.update(
                    dict(x for x in __resource.items()
                         if x[0] in {'headers', 'scheme', 'encoding', 'sample_size', 'allow_html',
                                     'force_strings', 'force_parse', 'skip_rows', 'compression'}))
                if isinstance(_params.get('skip_rows'), int):  # Backwards compatibility
                    _params['skip_rows'] = list(range(1, _params.get('skip_rows') + 1))

            if format is not None:
                _params['format'] = format

            if http_headers:
                http_session = requests.Session()
                http_session.headers = http_headers
                _params['http_session'] = http_session

            constants = _resource.get('constants', {})
            constant_headers = list(constants.keys())
            constant_values = [constants.get(k) for k in constant_headers]
            _stream = tabulator.Stream(__url, **_params,
                                       post_parse=[suffix_remover(format),
                                                   add_constants(constant_headers, constant_values, _columns)])
            retry = 0
            backoff = 2
            while True:
                try:
                    _stream.open()
                    _headers = dedupe(_stream.headers)
                    __columns = len(_headers)
                    _headers = dedupe(_headers + constant_headers)
                    _schema = __resource.get('schema')
                    if _schema is not None:
                        _schema = Schema(_schema)
                    return _schema, _headers, __columns, _stream, _stream.close
                except tabulator.exceptions.TabulatorException as e:
                    logging.warning("Error while opening resource from url %s: %r",
                                    _url, e)
                    _stream.close()
                    retry += 1
                    if retry <= 3:
                        logging.warning("Retrying after %d seconds (%d/3)", backoff, retry)
                        time.sleep(backoff)
                        backoff *= 2
                        continue
                    else:
                        if not _ignore_missing:
                            raise
                        return {}, [], 0, [], lambda: None
        return opener

    schema, headers, columns, stream, close = get_opener(url, _resource)()
    if schema is None:
        schema = {
            'fields': [
                {'name': header, 'type': 'string'}
                for header in headers
                ]
        }
        _resource['schema'] = schema

    close()
    del stream

    return itertools\
        .islice(
            _reader(
                get_opener(_url, _resource, columns),
                _url,
                max_row=limit_rows),
            1, None)


parameters, datapackage, resource_iterator = ingest()

resources = ResourceMatcher(parameters.get('resources'), datapackage)
ignore_missing = parameters.get('ignore-missing', False)
limit_rows = parameters.get('limit-rows', -1)

new_resource_iterator = []
for resource in datapackage['resources']:

    if streamable(resource):
        url = resource[PROP_STREAMED_FROM]

        name = resource['name']
        if not resources.match(name):
            continue

        path = get_path(resource)
        if path is None or path == PATH_PLACEHOLDER:
            path = os.path.join('data', name + '.csv')
            resource['path'] = path

        resource[PROP_STREAMING] = True

        rows = stream_reader(resource, url, ignore_missing or url == "", limit_rows,
                             resource.pop('http_headers', None))

        new_resource_iterator.append(rows)

    elif streaming(resource):
        new_resource_iterator.append(next(resource_iterator))

spew(datapackage, new_resource_iterator)



from dataflows import Flow, concatenate, update_resource
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.resources import PROP_STREAMING
from datapackage_pipelines.utilities.flow_utils import spew_flow


def flow(parameters):
    return Flow(
        concatenate(
            parameters.get('fields', {}),
            parameters.get('target', {}),
            parameters.get('sources')
        ),
        update_resource(
            parameters.get('target', {}).get('name', 'concat'),
            **{
                PROP_STREAMING: True
            }
        )
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



import os, logging
from datapackage_pipelines.utilities.lib_test_helpers import ProcessorFixtureTestsBase
from sqlalchemy.orm import sessionmaker
from sqlalchemy import create_engine, text

ROOT_PATH = os.path.join(os.path.dirname(__file__), '..', '..')
ENV = os.environ.copy()
ENV['PYTHONPATH'] = ROOT_PATH

ENV['EXISTENT_ENV'] = 'tests/data/sample.csv'

DEFAULT_TEST_DB = "sqlite://"
ENV['DPP_DB_ENGINE'] = os.environ.get("OVERRIDE_TEST_DB", DEFAULT_TEST_DB)


class StdlibfixtureTests(ProcessorFixtureTestsBase):

    def _get_procesor_env(self, filename):
        if ENV['DPP_DB_ENGINE'] != DEFAULT_TEST_DB:
            engine = create_engine(ENV['DPP_DB_ENGINE'])
            engine.execute(text("DROP TABLE IF EXISTS test;"))
        if filename == "dump_to_sql_update_mode__update":
            engine = create_engine(ENV['DPP_DB_ENGINE'])
            engine.execute(text("""
                CREATE TABLE test (
                  id integer not null primary key,
                  mystring text,
                  mynumber double precision,
                  mydate date
                )
            """))
            engine.execute(text("""
                INSERT INTO test VALUES (1, 'foo', 5.6, null);
            """))
        return ENV

    def _get_processor_file(self, processor):
        processor = processor.replace('.', '/')
        return os.path.join(ROOT_PATH, 'datapackage_pipelines', 'lib', processor.strip() + '.py')


for filename, _func in StdlibfixtureTests(os.path.join(os.path.dirname(__file__), 'fixtures')).get_tests():
    globals()['test_stdlib_%s' % filename] = _func




from dataflows import Flow, load
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow
from datapackage_pipelines.utilities.resources import PROP_STREAMING, PROP_STREAMED_FROM


def flow(parameters):
    _from = parameters.pop('from')

    num_resources = 0

    def count_resources():
        def func(package):
            global num_resources
            num_resources = len(package.pkg.resources)
            yield package.pkg
            yield from package
        return func

    def mark_streaming(_from):
        def func(package):
            for i in range(num_resources, len(package.pkg.resources)):
                package.pkg.descriptor['resources'][i].setdefault(PROP_STREAMING, True)
                package.pkg.descriptor['resources'][i].setdefault(PROP_STREAMED_FROM,  _from)
            yield package.pkg
            yield from package
        return func

    return Flow(
        count_resources(),
        load(_from, **parameters),
        mark_streaming(_from),
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



import collections


class LazyDict(collections.MutableMapping):

    def __init__(self):
        self._inner = None
        self._dirty = False

    @property
    def dirty(self):
        return self._dirty

    @property
    def inner(self):
        self.__ensure()
        return self._inner

    def _evaluate(self):
        raise NotImplementedError()

    def __ensure(self):
        if self._inner is None:
            self._inner = self._evaluate()

    def __len__(self):
        self.__ensure()
        return len(self._inner)

    def __getitem__(self, item):
        self.__ensure()
        return self._inner.__getitem__(item)

    def __setitem__(self, key, value):
        self.__ensure()
        self._inner.__setitem__(key, value)
        self._dirty = True

    def __delitem__(self, key):
        self.__ensure()
        self._inner.__delitem__(key)
        self._dirty = True

    def __iter__(self):
        self.__ensure()
        return self._inner.__iter__()



def is_a_url(path):
    return (path is not None and isinstance(path, str) and
            (path.startswith('http://') or
             path.startswith('https://'))
            )


def tabular(descriptor):
    return 'schema' in descriptor


def streaming(descriptor):
    return descriptor.get(PROP_STREAMING)


def streamable(descriptor):
    return PROP_STREAMED_FROM in descriptor and \
           not streaming(descriptor)


def get_path(descriptor):
    path = descriptor.get('path')
    if isinstance(path, str):
        return path
    if isinstance(path, list):
        if len(path) > 0:
            return path.pop(0)
        else:
            return None
    assert path is None, '%r' % path
    return None


PATH_PLACEHOLDER = '_'
PROP_STREAMED_FROM = 'dpp:streamedFrom'
PROP_STREAMING = 'dpp:streaming'



# -*- coding: utf-8 -*-
# Source: https://github.com/tsileo/dirtools (copied here because pypi package is not updated)

import logging
import os
import hashlib
from contextlib import closing  # for Python2.6 compatibility
import tarfile
import tempfile
from datetime import datetime
import json

from globster import Globster

log = logging.getLogger("dirtools")

# TODO abs=True args for .files(), .subdirs() ?


def load_patterns(exclude_file=".exclude"):
    """ Load patterns to exclude file from `exclude_file',
    and return a list of pattern.

    :type exclude_file: str
    :param exclude_file: File containing exclude patterns

    :rtype: list
    :return: List a patterns

    """
    return filter(None, open(exclude_file).read().split("\n"))


def _filehash(filepath, blocksize=4096):
    """ Return the hash object for the file `filepath', processing the file
    by chunk of `blocksize'.

    :type filepath: str
    :param filepath: Path to file

    :type blocksize: int
    :param blocksize: Size of the chunk when processing the file

    """
    sha = hashlib.sha256()
    with open(filepath, 'rb') as fp:
        while 1:
            data = fp.read(blocksize)
            if data:
                sha.update(data)
            else:
                break
    return sha


def filehash(filepath, blocksize=4096):
    """ Return the hash hexdigest() for the file `filepath', processing the file
    by chunk of `blocksize'.

    :type filepath: str
    :param filepath: Path to file

    :type blocksize: int
    :param blocksize: Size of the chunk when processing the file

    """
    sha = _filehash(filepath, blocksize)
    return sha.hexdigest()


class File(object):
    def __init__(self, path):
        self.file = os.path.basename(path)
        self.path = os.path.abspath(path)

    def _hash(self):
        """ Return the hash object. """
        return _filehash(self.path)

    def hash(self):
        """ Return the hash hexdigest. """
        return filehash(self.path)

    def compress_to(self, archive_path=None):
        """ Compress the directory with gzip using tarlib.

        :type archive_path: str
        :param archive_path: Path to the archive, if None, a tempfile is created

        """
        if archive_path is None:
            archive = tempfile.NamedTemporaryFile(delete=False)
            tar_args = ()
            tar_kwargs = {'fileobj': archive}
            _return = archive.name
        else:
            tar_args = (archive_path)
            tar_kwargs = {}
            _return = archive_path
        tar_kwargs.update({'mode': 'w:gz'})
        with closing(tarfile.open(*tar_args, **tar_kwargs)) as tar:
            tar.add(self.path, arcname=self.file)

        return _return


class Dir(object):
    """ Wrapper for dirtools arround a path.

    Try to load a .exclude file, ready to compute hashdir,


    :type directory: str
    :param directory: Root directory for initialization

    :type exclude_file: str
    :param exclude_file: File containing exclusion pattern,
        .exclude by default, you can also load .gitignore files.

    :type excludes: list
    :param excludes: List of additionals patterns for exclusion,
        by default: ['.git/', '.hg/', '.svn/']

    """
    def __init__(self, directory=".", exclude_file=".exclude",
                 excludes=['.git/', '.hg/', '.svn/']):
        if not os.path.isdir(directory):
            raise TypeError("Directory must be a directory.")
        self.directory = os.path.basename(directory)
        self.path = os.path.abspath(directory)
        self.parent = os.path.dirname(self.path)
        self.exclude_file = os.path.join(self.path, exclude_file)
        self.patterns = excludes
        if os.path.isfile(self.exclude_file):
            self.patterns.extend(load_patterns(self.exclude_file))
        self.globster = Globster(self.patterns)

    def hash(self, index_func=os.path.getmtime):
        """ Hash for the entire directory (except excluded files) recursively.

        Use mtime instead of sha256 by default for a faster hash.

        >>> dir.hash(index_func=dirtools.filehash)

        """
        # TODO alternative to filehash => mtime as a faster alternative
        shadir = hashlib.sha256()
        for f in self.files():
            try:
                shadir.update(str(index_func(os.path.join(self.path, f))))
            except (IOError, OSError):
                pass
        return shadir.hexdigest()

    def iterfiles(self, pattern=None, abspath=False):
        """ Generator for all the files not excluded recursively.

        Return relative path.

        :type pattern: str
        :param pattern: Unix style (glob like/gitignore like) pattern

        """
        if pattern is not None:
            globster = Globster([pattern])
        for root, dirs, files in self.walk():
            for f in files:
                if pattern is None or (pattern is not None and globster.match(f)):
                    if abspath:
                        yield os.path.join(root, f)
                    else:
                        yield self.relpath(os.path.join(root, f))

    def files(self, pattern=None, sort_key=lambda k: k, sort_reverse=False, abspath=False):
        """ Return a sorted list containing relative path of all files (recursively).

        :type pattern: str
        :param pattern: Unix style (glob like/gitignore like) pattern

        :param sort_key: key argument for sorted

        :param sort_reverse: reverse argument for sorted

        :rtype: list
        :return: List of all relative files paths.

        """
        return sorted(self.iterfiles(pattern, abspath=abspath), key=sort_key, reverse=sort_reverse)

    def get(self, pattern, sort_key=lambda k: k, sort_reverse=False, abspath=False):
        res = self.files(pattern, sort_key=sort_key, sort_reverse=sort_reverse, abspath=abspath)
        if res:
            return res[0]

    def itersubdirs(self, pattern=None, abspath=False):
        """ Generator for all subdirs (except excluded).

        :type pattern: str
        :param pattern: Unix style (glob like/gitignore like) pattern

        """
        if pattern is not None:
            globster = Globster([pattern])
        for root, dirs, files in self.walk():
            for d in dirs:
                if pattern is None or (pattern is not None and globster.match(d)):
                    if abspath:
                        yield os.path.join(root, d)
                    else:
                        yield self.relpath(os.path.join(root, d))

    def subdirs(self, pattern=None, sort_key=lambda k: k, sort_reverse=False, abspath=False):
        """ Return a sorted list containing relative path of all subdirs (recursively).

        :type pattern: str
        :param pattern: Unix style (glob like/gitignore like) pattern

        :param sort_key: key argument for sorted

        :param sort_reverse: reverse argument for sorted

        :rtype: list
        :return: List of all relative files paths.
        """
        return sorted(self.itersubdirs(pattern, abspath=abspath), key=sort_key, reverse=sort_reverse)

    def size(self):
        """ Return directory size in bytes.

        :rtype: int
        :return: Total directory size in bytes.
        """
        dir_size = 0
        for f in self.iterfiles(abspath=True):
            dir_size += os.path.getsize(f)
        return dir_size

    def is_excluded(self, path):
        """ Return True if `path' should be excluded
        given patterns in the `exclude_file'. """
        match = self.globster.match(self.relpath(path))
        if match:
            log.debug("{0} matched {1} for exclusion".format(path, match))
            return True
        return False

    def walk(self):
        """ Walk the directory like os.path
        (yields a 3-tuple (dirpath, dirnames, filenames)
        except it exclude all files/directories on the fly. """
        for root, dirs, files in os.walk(self.path, topdown=True):
            # TODO relative walk, recursive call if root excluder found???
            # root_excluder = get_root_excluder(root)
            ndirs = []
            # First we exclude directories
            for d in list(dirs):
                if self.is_excluded(os.path.join(root, d)):
                    dirs.remove(d)
                elif not os.path.islink(os.path.join(root, d)):
                    ndirs.append(d)

            nfiles = []
            for fpath in (os.path.join(root, f) for f in files):
                if not self.is_excluded(fpath) and not os.path.islink(fpath):
                    nfiles.append(os.path.relpath(fpath, root))

            yield root, ndirs, nfiles

    def find_projects(self, file_identifier=".project"):
        """ Search all directory recursively for subdirs
        with `file_identifier' in it.

        :type file_identifier: str
        :param file_identifier: File identier, .project by default.

        :rtype: list
        :return: The list of subdirs with a `file_identifier' in it.

        """
        projects = []
        for d in self.subdirs():
            project_file = os.path.join(self.directory, d, file_identifier)
            if os.path.isfile(project_file):
                projects.append(d)
        return projects

    def relpath(self, path):
        """ Return a relative filepath to path from Dir path. """
        return os.path.relpath(path, start=self.path)

    def compress_to(self, archive_path=None):
        """ Compress the directory with gzip using tarlib.

        :type archive_path: str
        :param archive_path: Path to the archive, if None, a tempfile is created

        """
        if archive_path is None:
            archive = tempfile.NamedTemporaryFile(delete=False)
            tar_args = []
            tar_kwargs = {'fileobj': archive}
            _return = archive.name
        else:
            tar_args = [archive_path]
            tar_kwargs = {}
            _return = archive_path
        tar_kwargs.update({'mode': 'w:gz'})
        with closing(tarfile.open(*tar_args, **tar_kwargs)) as tar:
            tar.add(self.path, arcname='', exclude=self.is_excluded)

        return _return


class DirState(object):
    """ Hold a directory state / snapshot meta-data for later comparison. """
    def __init__(self, _dir=None, state=None, index_cmp=os.path.getmtime):
        self._dir = _dir
        self.index_cmp = index_cmp
        self.state = state or self.compute_state()

    def compute_state(self):
        """ Generate the index. """
        data = {}
        data['directory'] = self._dir.path
        data['files'] = list(self._dir.files())
        data['subdirs'] = list(self._dir.subdirs())
        data['index'] = self.index()
        return data

    def index(self):
        index = {}
        for f in self._dir.iterfiles():
            try:
                index[f] = self.index_cmp(os.path.join(self._dir.path, f))
            except Exception as exc:
                print(f, exc)
        return index

    def __sub__(self, other):
        """ Compute diff with operator overloading.

        >>> path = DirState(Dir('/path'))
        >>> path_copy = DirState(Dir('/path_copy'))
        >>> diff =  path_copy - path
        >>> # Equals to
        >>> diff = compute_diff(path_copy.state, path.state)

        """
        if self.index_cmp != other.index_cmp:
            raise Exception('Both DirState instance must have the same index_cmp.')
        return compute_diff(self.state, other.state)

    def to_json(self, base_path='.', dt=None, fmt=None):
        if fmt is None:
            fmt = '{0}@{1}.json'
        if dt is None:
            dt = datetime.utcnow()
        path = fmt.format(self._dir.path.strip('/').split('/')[-1],
                          dt.isoformat())
        path = os.path.join(base_path, path)

        with open(path, 'wb') as f:
            f.write(json.dumps(self.state))

        return path

    @classmethod
    def from_json(cls, path):
        with open(path, 'rb') as f:
            return cls(state=json.loads(f.read()))


def compute_diff(dir_base, dir_cmp):
    """ Compare `dir_base' and `dir_cmp' and returns a list with
    the following keys:
     - deleted files `deleted'
     - created files `created'
     - updated files `updated'
     - deleted directories `deleted_dirs'

    """
    data = {}
    data['deleted'] = list(set(dir_cmp['files']) - set(dir_base['files']))
    data['created'] = list(set(dir_base['files']) - set(dir_cmp['files']))
    data['updated'] = []
    data['deleted_dirs'] = list(set(dir_cmp['subdirs']) - set(dir_base['subdirs']))

    for f in set(dir_cmp['files']).intersection(set(dir_base['files'])):
        if dir_base['index'][f] != dir_cmp['index'][f]:
            data['updated'].append(f)

    return data



def arg_to_step(arg):
    if isinstance(arg, str):
        return {'run': arg}
    else:
        return dict(zip(['run', 'parameters', 'cache'], arg))


def steps(*args):
    return [arg_to_step(arg) for arg in args]



import os

from celery.schedules import crontab

from .celery_common import get_celery_app, MANAGEMENT_TASK_NAME, SCHEDULED_TASK_NAME
from .celery_tasks import build_dependents
from datapackage_pipelines.specs import pipelines
from datapackage_pipelines.status import status_mgr

import logging

kw = {}
if os.environ.get('SCHEDULER'):
    CELERY_SCHEDULE = {
        '/management': {
            'task': MANAGEMENT_TASK_NAME,
            'schedule': crontab(),
            'args': ('update', None, None),
            'options': {'queue': 'datapackage-pipelines-management'}
        }
    }

    for spec in pipelines():
        if spec.schedule is not None:
            entry = {
                'task': SCHEDULED_TASK_NAME,
                'schedule': crontab(*spec.schedule),
                'args': (spec.pipeline_id,),
                'options': {'queue': 'datapackage-pipelines-management'}
            }
            CELERY_SCHEDULE[spec.pipeline_id] = entry
            logging.info('SCHEDULING task %r: %r', spec.pipeline_id, spec.schedule)

        ps = status_mgr().get(spec.pipeline_id)
        ex = ps.get_last_execution()
        if ex is not None and not ex.finish_time:
            ex.invalidate()
            ex.finish_execution(False, {}, ['Cancelled'])

    kw = dict(CELERYBEAT_SCHEDULE=CELERY_SCHEDULE)

logging.error('CELERY INITIALIZING')
celery_app = get_celery_app(**kw)

if os.environ.get('SCHEDULER'):
    build_dependents()
    celery_app.send_task(MANAGEMENT_TASK_NAME, ('init', None, None))



import datapackage
from datapackage.exceptions import DataPackageException
from ..parsers.base_parser import PipelineSpec

from ..errors import SpecError


class DependencyMissingException(Exception):

    def __init__(self, spec, missing):
        self.spec = spec
        self.missing = missing


def resolve_dependencies(spec: PipelineSpec, all_pipeline_ids, status_mgr):

    cache_hash = ''
    dependencies = spec.pipeline_details.get('dependencies', ())
    for dependency in dependencies:
        if 'pipeline' in dependency:
            pipeline_id = dependency['pipeline']
            if pipeline_id not in all_pipeline_ids:
                raise DependencyMissingException(spec, pipeline_id)

    for dependency in dependencies:
        if 'pipeline' in dependency:
            pipeline_id = dependency['pipeline']
            ps = status_mgr.get(pipeline_id)
            if not ps.runnable():
                spec.validation_errors.append(
                    SpecError('Invalid dependency',
                              'Cannot run until dependency passes validation: {}'.format(pipeline_id))
                )
            elif ps.dirty():
                spec.validation_errors.append(
                    SpecError('Dirty dependency',
                              'Cannot run until dependency is executed: {}'.format(pipeline_id))
                )
            elif ps.get_last_execution() is not None and not ps.get_last_execution().success:
                spec.validation_errors.append(
                    SpecError('Dependency unsuccessful',
                              'Cannot run until dependency "{}" is successfully '
                              'executed'.format(pipeline_id))
                )

            for dep_err in ps.validation_errors:
                spec.validation_errors.append(
                    SpecError('From {}'.format(pipeline_id), dep_err)
                )

            pipeline_hash = all_pipeline_ids.get(pipeline_id).cache_hash
            assert pipeline_hash is not None
            cache_hash += pipeline_hash

            spec.dependencies.append(pipeline_id)

        elif 'datapackage' in dependency:
            dp_id = dependency['datapackage']
            try:
                dp = datapackage.DataPackage(dp_id)
                if 'hash' in dp.descriptor:
                    cache_hash += dp.descriptor['hash']
                else:
                    spec.validation_errors.append(
                        SpecError('Missing dependency',
                                  "Couldn't get data from datapackage %s"
                                  % dp_id))
            except DataPackageException:
                spec.validation_errors.append(
                    SpecError('Missing dependency',
                              "Couldn't open datapackage %s"
                              % dp_id))

        else:
            spec.validation_errors.append(
                SpecError('Missing dependency',
                          'Unknown dependency provided (%r)' % dependency))

    return cache_hash



from dataflows import Flow, find_replace
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow


def flow(parameters):
    return Flow(
        find_replace(
            parameters.get('fields', []),
            resources=parameters.get('resources')
        )
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



from dataflows import Flow, sort_rows
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow, load_lazy_json


def flow(parameters):
    return Flow(
        load_lazy_json(parameters.get('resources')),
        sort_rows(
            parameters['sort-by'],
            resources=parameters.get('resources'),
            reverse=parameters.get('reverse')
        )
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



import os
import json
import jsonschema

from ..errors import SpecError


schema_filename = 'pipeline-spec.schema.json'
schema_filename = os.path.join(os.path.dirname(__file__),
                               schema_filename)
schema = json.load(open(schema_filename))
validator = jsonschema.validators.validator_for(schema)
schema = validator(schema)


def validate_pipeline(pipeline_details, errors):
    try:
        schema.validate(pipeline_details)
    except jsonschema.ValidationError as e:
        errors.append(SpecError('Invalid Pipeline', str(e)))
        return False
    return True



import sys

from datapackage_pipelines.wrapper import ingest, spew

SINK_MAGIC = '>>> PROCESSED ROWS: '


def sink(res_iter_):
    count = 0
    for res in res_iter_:
        for row in res:
            count += 1
            if count % 100 == 0:
                sys.stderr.write('%s%d\n' % (SINK_MAGIC, count))
                sys.stderr.flush()
    sys.stderr.write('%s%d\n' % (SINK_MAGIC, count))
    sys.stderr.flush()
    yield from ()


if __name__ == '__main__':
    sys.stderr.write('%s%d\n' % (SINK_MAGIC, 0))
    sys.stderr.flush()
    params, dp, res_iter = ingest()
    spew({'name': 'boop', 'resources': []},
        sink(res_iter))



import sys
import copy
import logging

import datapackage
from tableschema.exceptions import ValidationError, CastError
from tableschema import Schema

from ..utilities.resources import PATH_PLACEHOLDER, streaming
from ..utilities.extended_json import json


class ResourceIterator(object):

    def __init__(self, infile, spec, orig_spec,
                 validate=False, debug=False):
        self.spec = spec
        self.table_schema = Schema(orig_spec['schema'])
        self.field_names = [f['name'] for f in orig_spec['schema']['fields']]
        self.validate = validate
        self.infile = infile
        self.debug = debug
        self.stopped = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.stopped:
            raise StopIteration()
        if self.debug:
            logging.error('WAITING')
        line = self.infile.readline().strip()
        if self.debug:
            logging.error('INGESTING: %r', line)
        if line == '':
            self.stopped = True
            raise StopIteration()
        line = json.loadl(line)
        if self.validate:
            to_validate = [line.get(f) for f in self.field_names]
            try:
                self.table_schema.cast_row(to_validate)
            except CastError as e:
                logging.error('Failed to validate row: %s', e)
                for i, err in enumerate(e.errors):
                    logging.error('%d) %s', i+1, err.message)
                raise ValueError('Casting failed for row %r' % line) from e
            except TypeError as e:
                raise ValueError('Validation failed for row %r' % line) from e

        return line

    def next(self):
        return self.__next__()


def read_json(infile, proxy=False):
    line_json = infile.readline().strip()
    if line_json == '':
        sys.exit(-3)
    if proxy:
        print(line_json)
    try:
        return json.loads(line_json)
    except json.JSONDecodeError:
        logging.exception("Failed to decode line %r\nPerhaps there's a rogue print statement somewhere?", line_json)


def process_input(infile, validate=False, debug=False):
    dependency_dp = read_json(infile, True)
    dp = read_json(infile)
    resources = dp.get('resources', [])
    original_resources = copy.deepcopy(resources)

    if len(dp.get('resources', [])) == 0:
        # Currently datapackages with no resources are disallowed in the schema.
        # Since this might happen in the early stages of a pipeline,
        # we're adding this hack to avoid validation errors
        dp_to_validate = copy.deepcopy(dp)
        dp_to_validate['resources'] = [{
            'name': '__placeholder__',
            'path': PATH_PLACEHOLDER
        }]
    else:
        dp_to_validate = dp
    try:
        datapackage.validate(dp_to_validate)
    except ValidationError as e:
        logging.info('FAILED TO VALIDATE %r', dp_to_validate)
        for e in e.errors:
            try:
                logging.error("Data Package validation error: %s at dp%s",
                              e.message,
                              "[%s]" % "][".join(repr(index) for index in e.path))
            except AttributeError:
                logging.error("Data Package validation error: %s", e)
        raise

    infile.readline().strip()

    def resources_iterator(_resources, _original_resources):
        # we pass a resource instance that may be changed by the processing
        # code, so we must keep a copy of the original resource (used to
        # validate incoming data)
        ret = []
        for resource, orig_resource in zip(_resources, _original_resources):
            if not streaming(resource):
                continue

            res_iter = ResourceIterator(infile,
                                        resource, orig_resource,
                                        validate, debug)
            ret.append(res_iter)
        return iter(ret)

    return dp, resources_iterator(resources, original_resources), dependency_dp



import os

from dataflows import Flow, dump_to_path
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow

from datapackage_pipelines.utilities.stat_utils import STATS_DPP_KEY, STATS_OUT_DP_URL_KEY


def flow(parameters: dict, stats: dict):
    out_path = parameters.pop('out-path', '.')
    stats.setdefault(STATS_DPP_KEY, {})[STATS_OUT_DP_URL_KEY] = os.path.join(out_path, 'datapackage.json')
    return Flow(
        dump_to_path(
            out_path,
            **parameters
        )
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters, ctx.stats), ctx)



import datapackage

from datapackage_pipelines.wrapper import ingest, spew, get_dependency_datapackage_url

dep_prefix = 'dependency://'

parameters, dp, res_iter = ingest()

url = parameters['url']
if url.startswith(dep_prefix):
    dependency = url[len(dep_prefix):].strip()
    url = get_dependency_datapackage_url(dependency)
    assert url is not None, "Failed to fetch output datapackage for dependency '%s'" % dependency

datapackage = datapackage.DataPackage(url)
for k, v in datapackage.descriptor.items():
    if k != 'resources':
        dp[k] = v

spew(dp, res_iter)



import os
import shutil
import warnings

from datapackage_pipelines.lib.dump.dumper_base import FileDumper


class PathDumper(FileDumper):

    def initialize(self, params):
        super(PathDumper, self).initialize(params)
        self.out_path = params.get('out-path', '.')
        self.add_filehash_to_path = params.get('add-filehash-to-path', False)
        PathDumper.__makedirs(self.out_path)

    def write_file_to_output(self, filename, path):
        path = os.path.join(self.out_path, path)
        # Avoid rewriting existing files
        if self.add_filehash_to_path and os.path.exists(path):
            return
        path_part = os.path.dirname(path)
        PathDumper.__makedirs(path_part)
        shutil.copy(filename, path)
        os.chmod(path, 0o666)
        return path

    @staticmethod
    def __makedirs(path):
        os.makedirs(path, exist_ok=True)


if __name__ == '__main__':
    warnings.warn(
        'dump.to_path will be removed in the future, use "dump_to_path" instead',
        DeprecationWarning
    )
    PathDumper()()



import sys
import json
import os

import click
import requests

from .utilities.stat_utils import user_facing_stats

from .manager.logging_config import logging

from .specs import pipelines, PipelineSpec #noqa
from .status import status_mgr
from .manager import run_pipelines


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        click.echo('Available Pipelines:')
        for spec in pipelines():  # type: PipelineSpec
            ps = status_mgr().get(spec.pipeline_id)
            click.echo('- {} {}{}'
                       .format(spec.pipeline_id,
                               '(*)' if ps.dirty() else '',
                               '(E)' if len(spec.validation_errors) > 0 else ''))
            for error in spec.validation_errors:
                click.echo('\t{}: {}'.format(error.short_msg,
                                             error.long_msg))


@cli.command()
def serve():
    """Start the web server"""
    from .web import app
    app.run(host='0.0.0.0', debug=False, port=5000)


@cli.command()
@click.argument('pipeline_id')
@click.option('--verbose', default=False, is_flag=True)
@click.option('--use-cache/--no-use-cache', default=True,
              help='Cache (or don\'t) intermediate results (if requested in the pipeline)')
@click.option('--dirty', default=False, is_flag=True,
              help='Only run dirty pipelines')
@click.option('--force', default=False, is_flag=True)
@click.option('--concurrency', default=1)
@click.option('--slave', default=False, is_flag=True)
def run(pipeline_id, verbose, use_cache, dirty, force, concurrency, slave):
    """Run a pipeline by pipeline-id.
       pipeline-id supports '%' wildcard for any-suffix matching,
       'all' for running all pipelines and
       comma-delimited list of pipeline ids"""
    exitcode = 0

    running = []
    progress = {}

    def progress_cb(report):
        pid, count, success, *_, stats = report

        print('\x1b[%sA' % (1+len(running)))
        if pid not in progress:
            running.append(pid)
        progress[pid] = count, success

        for pid in running:
            count, success = progress[pid]
            if success is None:
                if count == 0:
                    print('\x1b[2K%s: \x1b[31m%s\x1b[0m' % (pid, 'WAITING FOR OUTPUT'))
                else:
                    print('\x1b[2K%s: \x1b[33mRUNNING, processed %s rows\x1b[0m' % (pid, count))
            else:
                if success:
                    print('\x1b[2K%s: \x1b[32mSUCCESS, processed %s rows\x1b[0m' % (pid, count))
                else:
                    print('\x1b[2K%s: \x1b[31mFAILURE, processed %s rows\x1b[0m' % (pid, count))

    results = run_pipelines(pipeline_id, '.', use_cache,
                            dirty, force, concurrency,
                            verbose, progress_cb if not verbose else None,
                            slave)
    if not slave:
        logging.info('RESULTS:')
        errd = False
        for result in results:
            stats = user_facing_stats(result.stats)
            errd = errd or result.errors or not result.success
            logging.info('%s: %s %s%s',
                         'SUCCESS' if result.success else 'FAILURE',
                         result.pipeline_id,
                         repr(stats) if stats is not None else '',
                         (
                            '\nERROR log from processor %s:\n+--------\n| ' % result.errors[0] +
                            '\n| '.join(result.errors[1:]) +
                            '\n+--------'
                         ) if result.errors else '')
    else:
        result_obj = []
        errd = False
        for result in results:
            errd = errd or result.errors or not result.success
            stats = user_facing_stats(result.stats)
            result_obj.append(dict(
                success=result.success,
                pipeline_id=result.pipeline_id,
                stats=result.stats,
                errors=result.errors
            ))
            json.dump(result_obj, sys.stderr)

    if errd:
        exitcode = 1

    exit(exitcode)


@cli.command()
def init():
    """Reset the status of all pipelines"""
    status_mgr().initialize()


@cli.command()
def version():
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'VERSION')) as f:
        installed = f.read().strip()
    latest = requests.get('https://pypi.org/pypi/datapackage-pipelines/json').json()['info']['version']
    print(f'Installed version: {installed}')
    print(f'Latest version: {latest}\n')
    if installed != latest:
        print('Datapackage Pipelines upgrade is available, upgrade using pip:\n')
        print('    python3 -m pip install -U datapackage-pipelines\n')
        exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    sys.exit(cli())
    # For Profiling:
    # import cProfile
    # sys.exit(cProfile.run('cli()', sort='cumulative'))



import os
import yaml

from .local_python import LocalPythonRunner, WrappedPythonRunner


class RunnerConfiguration(object):

    ENV_VAR = 'DPP_RUNNER_CONFIG'
    DEFAULT_RUNNER_CONFIG = 'dpp-runners.yaml'

    def __init__(self):

        config_fn = os.environ.get(self.ENV_VAR, self.DEFAULT_RUNNER_CONFIG)
        if os.path.exists(config_fn):
            self.config = yaml.load(open(config_fn))
        else:
            self.config = {}

    def get_runner_class(self, kind):
        return {
            'local-python': LocalPythonRunner,
            'wrapped-python': WrappedPythonRunner,
        }.get(kind, LocalPythonRunner)

    def get_runner(self, name):
        runner_config = self.config.get(name, {})
        kind = runner_config.get('kind')
        parameters = runner_config.get('parameters')
        return self.get_runner_class(kind)(name, parameters)



import asyncio
import gzip
import logging
import os
from concurrent.futures import CancelledError
from json.decoder import JSONDecodeError

from ..specs.specs import resolve_executor
from ..status import status_mgr
from ..utilities.stat_utils import STATS_DPP_KEY, STATS_OUT_DP_URL_KEY
from ..utilities.extended_json import json

from .runners import runner_config

SINK = os.path.join(os.path.dirname(__file__),
                    '..', 'lib', 'internal', 'sink.py')


async def enqueue_errors(step, process, queue, debug):
    out = process.stderr
    errors = []
    while True:
        try:
            line = await out.readline()
        except ValueError:
            logging.error('Received a too long log line (>64KB), discarded')
            continue
        if line == b'':
            break
        line = line.decode('utf8').rstrip()
        if len(line) != 0:
            if len(errors) == 0:
                if line.startswith('ERROR') or line.startswith('Traceback'):
                    errors.append(step['run'])
            if len(errors) > 0:
                errors.append(line)
                if len(errors) > 1000:
                    errors.pop(1)
            line = "{}: {}".format(step['run'], line)
            if debug:
                logging.info(line)
            await queue.put(line)
    return errors


async def dequeue_errors(queue, out):
    while True:
        line = await queue.get()
        if line is None:
            break
        out.append(line)
        if len(out) > 1000:
            out.pop(0)


async def collect_stats(infile):
    reader = asyncio.StreamReader()
    reader_protocol = asyncio.StreamReaderProtocol(reader)
    transport, _ = await asyncio.get_event_loop() \
        .connect_read_pipe(lambda: reader_protocol, infile)
    count = 0
    dp = None
    stats = None
    while True:
        try:
            line = await reader.readline()
        except ValueError:
            logging.exception('Too large stats object!')
            break
        if line == b'':
            break
        stats = line
        if dp is None:
            try:
                dp = json.loads(line.decode('ascii'))
            except JSONDecodeError:
                break
        count += 1

    transport.close()

    if dp is None or count == 0:
        return {}

    try:
        stats = json.loads(stats.decode('ascii'))
    except JSONDecodeError:
        stats = {}

    return stats


def create_process(args, cwd, wfd, rfd):
    pass_fds = {rfd, wfd}
    if None in pass_fds:
        pass_fds.remove(None)
    rfd = asyncio.subprocess.PIPE if rfd is None else rfd
    wfd = asyncio.subprocess.DEVNULL if wfd is None else wfd
    ret = asyncio.create_subprocess_exec(*args,
                                         stdin=rfd,
                                         stdout=wfd,
                                         stderr=asyncio.subprocess.PIPE,
                                         pass_fds=pass_fds,
                                         cwd=cwd)
    return ret


async def process_death_waiter(process):
    return_code = await process.wait()
    return process, return_code


def find_caches(pipeline_steps, pipeline_cwd):
    if not any(step.get('cache') for step in pipeline_steps):
        # If no step requires caching then bail
        return pipeline_steps

    for i, step in reversed(list(enumerate(pipeline_steps))):
        cache_filename = os.path.join(pipeline_cwd,
                                      '.cache',
                                      step['_cache_hash'])
        if os.path.exists(cache_filename):
            try:
                canary = gzip.open(cache_filename, "rt")
                canary.seek(1)
                canary.close()
            except Exception:  #noqa
                continue
            logging.info('Found cache for step %d: %s', i, step['run'])
            pipeline_steps = pipeline_steps[i+1:]
            step = {
                'run': 'cache_loader',
                'parameters': {
                    'load-from': os.path.join('.cache', step['_cache_hash'])
                }
            }
            step['executor'] = resolve_executor(step, '.', [])
            pipeline_steps.insert(0, step)
            break

    return pipeline_steps


async def construct_process_pipeline(pipeline_steps, pipeline_cwd, errors, debug=False):
    error_collectors = []
    processes = []
    error_queue = asyncio.Queue()
    rfd = None

    error_aggregator = \
        asyncio.ensure_future(dequeue_errors(error_queue, errors))

    pipeline_steps.append({
        'run': '(sink)',
        'executor': SINK,
        '_cache_hash': pipeline_steps[-1]['_cache_hash']
    })

    for i, step in enumerate(pipeline_steps):

        new_rfd, wfd = os.pipe()

        if debug:
            logging.info("- %s", step['run'])
        runner = runner_config.get_runner(step.get('runner'))
        args = runner.get_execution_args(step, pipeline_cwd, i)
        process = await create_process(args, pipeline_cwd, wfd, rfd)
        process.args = args[1]
        if wfd is not None:
            os.close(wfd)
        if rfd is not None:
            os.close(rfd)

        processes.append(process)
        rfd = new_rfd
        error_collectors.append(
            asyncio.ensure_future(enqueue_errors(step, process, error_queue, debug))
        )

    error_collectors.append(
        asyncio.ensure_future(collect_stats(os.fdopen(rfd)))
    )

    def wait_for_finish(_error_collectors,
                        _error_queue,
                        _error_aggregator):
        async def _func(failed_index=None):
            *errors, count = await asyncio.gather(*_error_collectors)
            if failed_index is not None:
                errors = errors[failed_index]
            else:
                errors = None
            await _error_queue.put(None)
            await _error_aggregator
            return count, errors
        return _func

    return processes, \
        wait_for_finish(error_collectors,
                        error_queue,
                        error_aggregator)


async def async_execute_pipeline(pipeline_id,
                                 pipeline_steps,
                                 pipeline_cwd,
                                 trigger,
                                 execution_id,
                                 use_cache,
                                 dependencies,
                                 debug):

    if debug:
        logging.info("%s Async task starting", execution_id[:8])
    ps = status_mgr().get(pipeline_id)
    if not ps.start_execution(execution_id):
        logging.info("%s START EXECUTION FAILED %s, BAILING OUT", execution_id[:8], pipeline_id)
        return False, {}, []

    ps.update_execution(execution_id, [])

    if use_cache:
        if debug:
            logging.info("%s Searching for existing caches", execution_id[:8])
        pipeline_steps = find_caches(pipeline_steps, pipeline_cwd)
    execution_log = []

    if debug:
        logging.info("%s Building process chain:", execution_id[:8])
    processes, stop_error_collecting = \
        await construct_process_pipeline(pipeline_steps, pipeline_cwd, execution_log, debug)

    processes[0].stdin.write(json.dumps(dependencies).encode('utf8') + b'\n')
    processes[0].stdin.write(b'{"name": "_", "resources": []}\n')
    processes[0].stdin.close()

    def kill_all_processes():
        for to_kill in processes:
            try:
                to_kill.kill()
            except ProcessLookupError:
                pass

    success = True
    pending = [asyncio.ensure_future(process_death_waiter(process))
               for process in processes]
    index_for_pid = dict(
        (p.pid, i)
        for i, p in enumerate(processes)
    )
    failed_index = None
    while len(pending) > 0:
        done = []
        try:
            done, pending = \
                await asyncio.wait(pending,
                                   return_when=asyncio.FIRST_COMPLETED,
                                   timeout=10)
        except CancelledError:
            success = False
            kill_all_processes()

        for waiter in done:
            process, return_code = waiter.result()
            if return_code == 0:
                if debug:
                    logging.info("%s DONE %s", execution_id[:8], process.args)
                processes = [p for p in processes if p.pid != process.pid]
            else:
                if return_code > 0 and failed_index is None:
                    failed_index = index_for_pid[process.pid]
                if debug:
                    logging.error("%s FAILED %s: %s", execution_id[:8], process.args, return_code)
                success = False
                kill_all_processes()

        if success and not ps.update_execution(execution_id, execution_log):
            logging.error("%s FAILED to update %s", execution_id[:8], pipeline_id)
            success = False
            kill_all_processes()

    stats, error_log = await stop_error_collecting(failed_index)
    if success is False:
        stats = None

    ps.update_execution(execution_id, execution_log, hooks=True)
    ps.finish_execution(execution_id, success, stats, error_log)

    logging.info("%s DONE %s %s %r", execution_id[:8], 'V' if success else 'X', pipeline_id, stats)

    return success, stats, error_log


def execute_pipeline(spec,
                     execution_id,
                     trigger='manual',
                     use_cache=True):

    debug = trigger == 'manual' or os.environ.get('DPP_DEBUG')
    logging.info("%s RUNNING %s", execution_id[:8], spec.pipeline_id)

    loop = asyncio.get_event_loop()

    if debug:
        logging.info("%s Collecting dependencies", execution_id[:8])
    dependencies = {}
    for dep in spec.pipeline_details.get('dependencies', []):
        if 'pipeline' in dep:
            dep_pipeline_id = dep['pipeline']
            pipeline_execution = status_mgr().get(dep_pipeline_id).get_last_successful_execution()
            if pipeline_execution is not None:
                result_dp = pipeline_execution.stats.get(STATS_DPP_KEY, {}).get(STATS_OUT_DP_URL_KEY)
                if result_dp is not None:
                    dependencies[dep_pipeline_id] = result_dp

    if debug:
        logging.info("%s Running async task", execution_id[:8])

    pipeline_task = \
        asyncio.ensure_future(async_execute_pipeline(spec.pipeline_id,
                                                     spec.pipeline_details.get('pipeline', []),
                                                     spec.path,
                                                     trigger,
                                                     execution_id,
                                                     use_cache,
                                                     dependencies,
                                                     debug))
    try:
        if debug:
            logging.info("%s Waiting for completion", execution_id[:8])
        return loop.run_until_complete(pipeline_task)
    except KeyboardInterrupt:
        logging.info("Caught keyboard interrupt. Cancelling tasks...")
        pipeline_task.cancel()
        loop.run_forever()
        logging.info("Caught keyboard interrupt. DONE!")
        raise KeyboardInterrupt()


def finalize():
    loop = asyncio.get_event_loop()
    loop.close()



from dataflows import Flow, dump_to_sql
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow


def flow(parameters):
    return Flow(
        dump_to_sql(
            parameters['tables'],
            engine=parameters.get('engine', 'env://DPP_DB_ENGINE'),
            updated_column=parameters.get("updated_column"),
            updated_id_column=parameters.get("updated_id_column")
        )
    )


if __name__ == '__main__':
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



import sys
import shutil
import gzip

from datapackage_pipelines.wrapper import ingest

params, _, _ = ingest()

load_from = params['load-from']

shutil.copyfileobj(gzip.open(load_from, "rt"), sys.stdout)



import logging
import os
import hashlib
from importlib.util import find_spec
from importlib import import_module

from .errors import SpecError

_processor_path = None


def processor_path():
    global _processor_path
    if _processor_path is None:
        _processor_path = os.environ.get('DPP_PROCESSOR_PATH', '').split(';')
    return _processor_path


_found_files = set()


def find_file_in_path(path, remove=0):
    def finder(parts):
        global _found_files
        filename = os.path.join(*(path + parts[remove:]))
        if filename in _found_files:
            return filename
        if os.path.exists(filename):
            _found_files.add(filename)
            return filename
    return finder


def convert_dot_notation(executor):
    parts = []
    back_up = False
    while executor.startswith('..'):
        parts.append('..')
        executor = executor[1:]
        back_up = True

    if executor.startswith('.'):
        executor = executor[1:]

    executor = executor.split('.')
    executor[-1] += '.py'

    parts.extend(executor)

    return back_up, parts


_tried_imports = {}


def load_module(module):
    global _tried_imports
    if module in _tried_imports:
        return _tried_imports[module]
    module_name = 'datapackage_pipelines_'+module
    ret = None
    if find_spec(module_name):
        ret = import_module(module_name)
    _tried_imports[module] = ret
    return ret


def resolve_executor(step, path, errors):

    if 'code' in step:
        filename = hashlib.md5(step['code'].encode('utf8')).hexdigest()
        code_path = os.path.join(path, '.code')
        if not os.path.exists(code_path):
            os.mkdir(code_path)
        code_path = os.path.join(code_path, filename)
        with open(code_path, 'w') as code_file:
            code_file.write(step['code'])
        return code_path

    if 'flow' in step:
        step['run'] = 'flow'
        parameters = step.setdefault('parameters', {})
        flow = step.pop('flow')
        parameters.update(__flow=flow, __path=path)
        try:
            m = import_module(flow)
            flow_py_file = m.__file__
            parameters['__flow_hash'] = hashlib.md5(open(flow_py_file, 'rb').read()).hexdigest()
        except Exception:
            pass

    executor = step['run']
    back_up, parts = convert_dot_notation(executor)
    resolvers = [find_file_in_path([path])]
    if not back_up:
        if len(parts) > 1:
            module_name = parts[0]
            module = load_module(module_name)
            if module is not None:
                module = list(module.__path__)[0]
                resolvers.append(find_file_in_path([module, 'processors'], 1))

        resolvers.extend([
            find_file_in_path([path_])
            for path_ in processor_path()
        ])
        resolvers.append(find_file_in_path([os.path.dirname(__file__),
                                            '..', 'lib']))

    for resolver in resolvers:
        location = resolver(parts)
        if location is not None:
            return location

    message = "Couldn't resolve {0} at {1}".format(executor, path)
    errors.append(SpecError('Unresolved processor', message))


resolved_generators = {}


def resolve_generator(module_name):
    if module_name in resolved_generators:
        return resolved_generators[module_name]
    resolved_generators[module_name] = None
    module = load_module(module_name)
    if module is None:
        return None
    try:
        generator_class = module.Generator
    except AttributeError:
        logging.warning("Can't find 'Generator' identifier in %s", module_name)
        return None
    generator = generator_class()
    resolved_generators[module_name] = generator
    return generator



from datapackage_pipelines.wrapper import ingest, spew
import os

from datapackage_pipelines.utilities.resources import PATH_PLACEHOLDER, PROP_STREAMED_FROM

parameters, datapackage, res_iter = ingest()


if datapackage is None:
    datapackage = {}

datapackage.setdefault('resources', [])

for param in ['url', 'name']:
    assert param in parameters, \
        "You must define {} in your parameters".format(param)

url = parameters.pop('url')
if url.startswith('env://'):
    env_var = url[6:]
    env_url = os.environ.get(env_var)
    assert env_url is not None, \
        "Missing Value - " \
        "Please set your '%s' environment variable" % env_var

    url = env_url

if 'path' not in parameters:
    parameters['path'] = PATH_PLACEHOLDER
parameters[PROP_STREAMED_FROM] = url

datapackage['resources'].append(parameters)

spew(datapackage, res_iter)



import sys
from importlib import import_module
from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow


with ingest() as ctx:
    parameters, datapackage, resources = ctx
    stats = {}

    sys.path.append(parameters.pop('__path'))
    flow_module = import_module(parameters.pop('__flow'))
    flow = flow_module.flow(parameters, datapackage, resources, ctx.stats)

    spew_flow(flow, ctx)



import os
import codecs
import ujson


class FilesystemBackend(object):

    KIND = 'filesystem'

    def __init__(self, root_dir='.'):
        dpp_dirname = os.environ.get('DPP_DB_DIRNAME', '.dpp')
        self.base_dir = os.path.join(root_dir, dpp_dirname)
        os.makedirs(self.base_dir, exist_ok=True)

    def fn(self, pipeline_id):
        pipeline_id = codecs.encode(pipeline_id.encode('utf8'), 'base64').decode('ascii').replace('\n', '')
        return os.path.join(self.base_dir, pipeline_id)

    def get_status(self, pipeline_id):
        try:
            with open(self.fn(pipeline_id)) as f:
                return ujson.load(f)
        except FileNotFoundError:
            pass
        except ValueError:
            pass

    def set_status(self, pipeline_id, status):
        fn = self.fn(pipeline_id)
        with open(fn+'.tmp', 'w') as f:
            ujson.dump(status, f)
        os.rename(fn+'.tmp', fn)

    def del_status(self, pipeline_id):
        try:
            os.unlink(self.fn(pipeline_id))
        except FileNotFoundError:
            pass

    def register_pipeline_id(self, pipeline_id):
        pass

    def deregister_pipeline_id(self, pipeline_id):
        self.del_status(pipeline_id)

    def reset(self):
        for p in self.all_pipeline_ids():
            self.del_status(p)

    def all_pipeline_ids(self):
        # Decoding encoded identifiers
        dec_ids = []
        enc_ids = sorted(os.listdir(self.base_dir))
        for enc_id in enc_ids:
            dec_id = codecs.decode(enc_id.encode('utf8'), 'base64').decode('utf8')
            if dec_id.startswith('PipelineStatus:'):
                dec_id = dec_id.replace('PipelineStatus:', '')
                dec_ids.append(dec_id)
        return dec_ids

    def all_statuses(self):
        return [self.get_status(_id)
                for _id in self.all_pipeline_ids()]



import os
from typing import Iterator  #noqa

import yaml
from datapackage_pipelines.status import status_mgr
from datapackage_pipelines.utilities import dirtools

from .resolver import resolve_executor
from .errors import SpecError
from .schemas.validator import validate_pipeline
from .parsers import BasicPipelineParser, SourceSpecPipelineParser
from .parsers.base_parser import PipelineSpec
from .hashers import HashCalculator, DependencyMissingException

SPEC_PARSERS = [
    BasicPipelineParser(),
    SourceSpecPipelineParser()
]

YAML_LOADER = yaml.CLoader if 'CLoader' in yaml.__dict__ else yaml.Loader


def resolve_processors(spec: PipelineSpec):
    abspath = os.path.abspath(spec.path)
    for step in spec.pipeline_details['pipeline']:
        if 'executor' not in step:
            step['executor'] = resolve_executor(step,
                                                abspath,
                                                spec.validation_errors)


def process_schedules(spec: PipelineSpec):
    if spec.schedule is None:
        schedule = spec.pipeline_details.get('schedule', {})
        if 'crontab' in schedule:
            schedule = schedule['crontab'].split()
            spec.schedule = schedule


def find_specs(root_dir='.') -> PipelineSpec:
    for dirpath, dirnames, filenames in dirtools.Dir(root_dir,
                                                     exclude_file='.dpp_spec_ignore',
                                                     excludes=['.*']).walk():
        relpath = os.path.relpath(dirpath, root_dir)
        dirpath = os.path.join(root_dir, relpath) if relpath != '.' else '.'
        if dirpath.startswith(os.path.join(root_dir, '.')):
            continue
        for filename in filenames:
            for parser in SPEC_PARSERS:
                if parser.check_filename(filename):
                    fullpath = os.path.join(dirpath, filename)
                    with open(fullpath, encoding='utf8') as spec_file:
                        contents = spec_file.read()
                        try:
                            spec = yaml.load(contents, Loader=YAML_LOADER)
                            yield from parser.to_pipeline(spec, fullpath, root_dir)
                        except yaml.YAMLError as e:
                            error = SpecError('Invalid Spec File %s' % fullpath, str(e))
                            yield PipelineSpec(path=dirpath, validation_errors=[error])


def pipelines(prefixes=None, ignore_missing_deps=False, root_dir='.', status_manager=None):

    specs: Iterator[PipelineSpec] = find_specs(root_dir)
    hasher = HashCalculator()
    if status_manager is None:
        status_manager = status_mgr()
    if prefixes is None:
        prefixes = ('',)
    while specs is not None:
        deferred = []
        found = False

        for spec_ in specs:
            spec: PipelineSpec = spec_

            if not any(spec.pipeline_id.startswith(prefix)
                       for prefix in prefixes):
                continue

            if (spec.pipeline_details is not None and
                    validate_pipeline(spec.pipeline_details, spec.validation_errors)):

                resolve_processors(spec)
                process_schedules(spec)

                try:
                    hasher.calculate_hash(spec, status_manager, ignore_missing_deps)
                    found = True
                except DependencyMissingException as e_:
                    e: DependencyMissingException = e_
                    deferred.append((e.spec, e.missing))
                    continue

            yield spec

        if found and len(deferred) > 0:
            specs = iter((x[0] for x in deferred))
        else:
            for spec, missing in deferred:
                spec.validation_errors.append(
                    SpecError('Missing dependency',
                              'Failed to find a dependency: {}'.format(missing))
                )
                yield spec
            specs = None


def register_all_pipelines(root_dir='.'):
    for spec in pipelines(root_dir=root_dir):
        ps = status_mgr().get(spec.pipeline_id)
        ps.init(spec.pipeline_details,
                spec.source_details,
                spec.validation_errors,
                spec.cache_hash)
        ps.save()



import hashlib

from ...utilities.extended_json import json
from ..parsers.base_parser import PipelineSpec

from ..errors import SpecError
from .dependency_resolver import resolve_dependencies


class HashCalculator(object):

    def __init__(self):
        self.all_pipeline_ids = {}

    def calculate_hash(self, spec: PipelineSpec, status_mgr, ignore_missing_deps=False):

        cache_hash = None
        if spec.pipeline_id in self.all_pipeline_ids:
            message = 'Duplicate key {0} in {1}' \
                .format(spec.pipeline_id, spec.path)
            spec.validation_errors.append(SpecError('Duplicate Pipeline Id', message))

        else:
            if ignore_missing_deps:
                cache_hash = ''
            else:
                cache_hash = resolve_dependencies(spec, self.all_pipeline_ids, status_mgr)

            self.all_pipeline_ids[spec.pipeline_id] = spec
            if len(spec.validation_errors) > 0:
                return cache_hash

            for step in spec.pipeline_details['pipeline']:
                m = hashlib.md5()
                m.update(cache_hash.encode('ascii'))
                with open(step['executor'], 'rb') as f:
                    m.update(f.read())
                m.update(json.dumps(step, ensure_ascii=True, sort_keys=True)
                         .encode('ascii'))
                cache_hash = m.hexdigest()
                step['_cache_hash'] = cache_hash

        spec.cache_hash = cache_hash



import itertools
import copy
import logging

import datapackage

from dataflows.helpers.resource_matcher import ResourceMatcher

from datapackage_pipelines.wrapper import ingest, spew, get_dependency_datapackage_url
from datapackage_pipelines.utilities.resources import tabular, PROP_STREAMING, \
    PROP_STREAMED_FROM


def progress_logger(iter, log_progress_rows):
    for i, row in enumerate(iter, 1):
        yield row
        if i % log_progress_rows == 0:
            logging.info('loaded {} rows'.format(i))


class ResourceLoader(object):

    def __init__(self):
        self.parameters, self.dp, self.res_iter = ingest()

    def __call__(self):
        url = self.parameters['url']
        limit_rows = self.parameters.get('limit-rows')
        log_progress_rows = self.parameters.get('log-progress-rows')
        dep_prefix = 'dependency://'
        if url.startswith(dep_prefix):
            dependency = url[len(dep_prefix):].strip()
            url = get_dependency_datapackage_url(dependency)
            assert url is not None, "Failed to fetch output datapackage for dependency '%s'" % dependency
        stream = self.parameters.get('stream', True)
        required = self.parameters.get('required', True)
        resource = self.parameters.get('resource')
        resources = self.parameters.get('resources')
        if resource is not None:
            assert not resources
            resource_index = resource if isinstance(resource, int) else None
        else:
            assert resources
            resource_index = None
            resource = list(resources.keys())
        name_matcher = (
            ResourceMatcher(resource, self.dp)
            if isinstance(resource, (str, list))
            else None
        )

        selected_resources = []
        found = False
        try:
            dp = datapackage.DataPackage(url)
        except Exception:
            if required:
                raise
            else:
                dp = None
        if dp:
            dp = self.process_datapackage(dp)
            for i, orig_res in enumerate(dp.resources):
                if resource_index == i or \
                        (name_matcher is not None and name_matcher.match(orig_res.descriptor.get('name'))):
                    found = True
                    desc = copy.deepcopy(orig_res.descriptor)
                    if 'primaryKey' in desc.get('schema', {}):
                        # Avoid duplication checks
                        del orig_res.descriptor['schema']['primaryKey']
                        orig_res.commit()
                    desc[PROP_STREAMED_FROM] = orig_res.source
                    if resources:
                        desc.update(resources[desc['name']])
                    self.dp['resources'].append(desc)
                    if tabular(desc) and stream:
                        desc[PROP_STREAMING] = True
                        orig_res_iter = orig_res.iter(keyed=True)
                        if limit_rows:
                            orig_res_iter = itertools.islice(orig_res_iter, limit_rows)
                        if log_progress_rows:
                            orig_res_iter = progress_logger(orig_res_iter, log_progress_rows)
                        selected_resources.append(orig_res_iter)
                    else:
                        desc[PROP_STREAMING] = False

        assert found or not required, "Failed to find resource with index or name matching %r" % resource
        spew(self.dp, itertools.chain(self.res_iter, selected_resources))

    def process_datapackage(self, dp_):
        return dp_


if __name__ == '__main__':
    ResourceLoader()()



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
from setuptools import setup, find_packages


# Helpers
def read(*paths):
    """Read a text file."""
    basedir = os.path.dirname(__file__)
    fullpath = os.path.join(basedir, *paths)
    contents = io.open(fullpath, encoding='utf-8').read().strip()
    return contents


# Prepare
PACKAGE = 'datapackage_pipelines'
NAME = PACKAGE.replace('_', '-')
INSTALL_REQUIRES = [
    'celery',
    'requests',
    'datapackage>=1.5.1',
    'tableschema>=1.2.5',
    'tableschema-sql>=0.10.4',
    'pyyaml',
    'ujson',
    'mistune',
    'redis>=3,<4',
    'click<8.0',
    'awesome-slugify',
    'flask<2.0.0',
    'flask-cors',
    'flask-jsonpify',
    'flask-basicauth',
    'cachetools',
    'tabulator>=1.17.0',
    'globster>=0.1.0',
    'dataflows>=0.0.57',
]
SPEEDUP_REQUIRES = [
    'dataflows[speedup]',
]
LINT_REQUIRES = [
    'pylama',
]
TESTS_REQUIRE = [
    'tox',
    'sqlalchemy',
]
README = read('README.md')
VERSION = read(PACKAGE, 'VERSION')
PACKAGES = find_packages(exclude=['examples', 'tests', '.tox'])

# Run
setup(
    name=NAME,
    version=VERSION,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require={
        'develop': LINT_REQUIRES + TESTS_REQUIRE,
        'speedup': SPEEDUP_REQUIRES,
    },
    zip_safe=False,
    long_description=README,
    long_description_content_type='text/markdown',
    description='{{ DESCRIPTION }}',
    author='Open Knowledge Foundation',
    author_email='info@okfn.org',
    url='https://github.com/frictionlessdata/datapackage-pipelines',
    license='MIT',
    keywords=[
        'data',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points={
      'console_scripts': [
        'dpp = datapackage_pipelines.cli:cli',
      ]
    },
)



import warnings

from datapackage_pipelines.wrapper import ingest
from datapackage_pipelines.utilities.flow_utils import spew_flow

from datapackage_pipelines.lib.update_package import flow


if __name__ == '__main__':
    warnings.warn(
        'add_metadata will be removed in the future, use "update_package" instead',
        DeprecationWarning
    )
    with ingest() as ctx:
        spew_flow(flow(ctx.parameters), ctx)



import datetime
import os
from io import BytesIO
import logging
from functools import wraps
from copy import deepcopy
from collections import Counter

import slugify
import yaml
import mistune
import requests

from flask import \
    Blueprint, Flask, render_template, abort, send_file, make_response
from flask_cors import CORS
from flask_jsonpify import jsonify
from flask_basicauth import BasicAuth

from datapackage_pipelines.status import status_mgr
from datapackage_pipelines.utilities.stat_utils import user_facing_stats

YAML_DUMPER = yaml.CDumper if 'CDumper' in yaml.__dict__ else yaml.Dumper


def datestr(x):
    if x is None:
        return ''
    return str(datetime.datetime.fromtimestamp(x))


def yamlize(x):
    ret = yaml.dump(x, default_flow_style=False, Dumper=YAML_DUMPER)
    return ret


markdown = mistune.Markdown(hard_wrap=True)
status = status_mgr()


def make_hierarchies(statuses):

    def group(l):
        pipelines = list(filter(lambda x: len(x['id']) == 1, l))
        children_ = list(filter(lambda x: len(x['id']) > 1, l))
        groups_ = {}
        for child in children_:
            child_key = child['id'].pop(0)
            groups_.setdefault(child_key, []).append(child)
        children_ = dict(
            (k, group(v))
            for k, v in groups_.items()
        )
        for p in pipelines:
            p['id'] = p['id'][0]
        return {
            'pipelines': pipelines,
            'children': children_
        }

    def flatten(children_):
        for k, v in children_.items():
            v['children'] = flatten(v['children'])
            child_keys = list(v['children'].keys())
            if len(child_keys) == 1 and len(v['pipelines']) == 0:
                child_key = child_keys[0]
                children_['/'.join([k, child_key])] = v['children'][child_key]
                del children_[k]
        return children_

    statuses = [
       {
            'id': st['id'].split('/'),
            'title': st.get('title'),
            'stats': st.get('stats'),
            'slug': st.get('slug')
       }
       for st in statuses
    ]
    groups = group(statuses)
    children = groups.get('children', {})
    groups['children'] = flatten(children)

    return groups


def basic_auth_required(view_func):
    """
    A decorator that can be used to protect specific views with HTTP basic
    access authentication. Conditional on having BASIC_AUTH_USERNAME and
    BASIC_AUTH_PASSWORD set as env vars.
    """
    @wraps(view_func)
    def wrapper(*args, **kwargs):
        if app.config.get('BASIC_AUTH_ACTIVE', False):
            if basic_auth.authenticate():
                return view_func(*args, **kwargs)
            else:
                return basic_auth.challenge()
        else:
            return view_func(*args, **kwargs)
    return wrapper


blueprint = Blueprint('dpp', 'dpp')


@blueprint.route("")
@blueprint.route("<path:pipeline_path>")
@basic_auth_required
def main(pipeline_path=None):
    pipeline_ids = sorted(status.all_pipeline_ids())

    # If we have a pipeline_path, filter the pipeline ids.
    if pipeline_path is not None:
        if not pipeline_path.startswith('./'):
            pipeline_path = './' + pipeline_path

        pipeline_ids = [p for p in pipeline_ids if p.startswith(pipeline_path)]

    statuses = []
    for pipeline_id in pipeline_ids:
        pipeline_status = status.get(pipeline_id)
        ex = pipeline_status.get_last_execution()
        success_ex = pipeline_status.get_last_successful_execution()
        pipeline_obj = {
            'id': pipeline_id.lstrip('./'),
            'title': pipeline_status.pipeline_details.get('title'),
            'stats': user_facing_stats(ex.stats) if ex else None,
            'slug': slugify.slugify(pipeline_id),
            'trigger': ex.trigger if ex else None,
            'error_log': pipeline_status.errors(),
            'state': pipeline_status.state(),
            'pipeline': pipeline_status.pipeline_details,
            'message': pipeline_status.state().capitalize(),
            'dirty': pipeline_status.dirty(),
            'runnable': pipeline_status.runnable(),
            'class': {'INIT': 'primary',
                      'QUEUED': 'primary',
                      'INVALID': 'danger',
                      'RUNNING': 'warning',
                      'SUCCEEDED': 'success',
                      'FAILED': 'danger'
                      }[pipeline_status.state()],
            'ended': datestr(ex.finish_time) if ex else None,
            'started': datestr(ex.start_time) if ex else None,
            'last_success':
                datestr(success_ex.finish_time) if success_ex else None,
        }
        statuses.append(pipeline_obj)

    def state_and_not_dirty(state, p):
        return p.get('state') == state and not p.get('dirty')

    def state_or_dirty(state, p):
        return p.get('state') == state or p.get('dirty')

    categories = [
        ['ALL', 'All Pipelines', lambda _, __: True],
        ['INVALID', "Can't start", lambda _, p: not p['runnable']],
        ['QUEUED', 'Waiting to run', lambda state, p: p['state'] == state],
        ['RUNNING', 'Running', state_and_not_dirty],
        ['FAILED', 'Failed Execution', state_and_not_dirty],
        ['SUCCEEDED', 'Successful Execution', state_and_not_dirty],
    ]
    for item in categories:
        item.append([p for p in deepcopy(statuses)
                     if item[2](item[0], p)])
        item.append(len(item[-1]))
        item.append(make_hierarchies(item[-2]))
    return render_template('dashboard.html',
                           categories=categories,
                           yamlize=yamlize,
                           markdown=markdown)


@blueprint.route("api/raw/status")
@basic_auth_required
def pipeline_raw_api_status():
    pipelines = sorted(status.all_statuses(), key=lambda x: x.get('id'))
    for pipeline in pipelines:
        # can get the full details from api/raw/<path:pipeline_id>
        for attr in ["pipeline", "reason", "error_log"]:
            if attr in pipeline:
                del pipeline[attr]
    return jsonify(pipelines)


@blueprint.route("api/raw/<path:pipeline_id>")
@basic_auth_required
def pipeline_raw_api(pipeline_id):
    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get(pipeline_id)
    if not pipeline_status.pipeline_details:
        abort(404)
    last_execution = pipeline_status.get_last_execution()
    last_successful_execution = pipeline_status.get_last_successful_execution()
    ret = {
        "id": pipeline_id,
        "cache_hash": pipeline_status.cache_hash,
        "dirty": pipeline_status.dirty(),

        "queued": last_execution.queue_time if last_execution else None,
        "started": last_execution.start_time if last_execution else None,
        "ended": last_execution.finish_time if last_execution else None,
        "reason": last_execution.log if last_execution else None,
        "error_log": pipeline_status.errors(),
        "stats": last_execution.stats if last_execution else None,
        "success": last_execution.success if last_execution else None,
        "last_success":
            last_successful_execution.finish_time
            if last_successful_execution else None,
        "trigger": last_execution.trigger if last_execution else None,

        "pipeline": pipeline_status.pipeline_details,
        "source": pipeline_status.source_spec,
        "message": pipeline_status.state().capitalize(),
        "state": pipeline_status.state(),
    }

    return jsonify(ret)


@blueprint.route("api/<field>/<path:pipeline_id>")
@basic_auth_required
def pipeline_api(field, pipeline_id):

    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get(pipeline_id)
    if not pipeline_status.pipeline_details:
        abort(404)

    ret = None
    if field == 'pipeline':
        ret = pipeline_status.pipeline_details
        ret = yamlize(ret)
    elif field == 'source':
        ret = pipeline_status.source_spec
        ret = yamlize(ret)
    elif field == 'log':
        ex = pipeline_status.get_last_execution()
        ret = ex.log if ex else ''
    else:
        abort(400)

    ret = ret.split('\n')
    ret = {'text': ret}
    return jsonify(ret)


def _make_badge_response(subject, text, colour):
    image_url = 'https://img.shields.io/badge/{}-{}-{}.svg'.format(
        subject, text, colour)
    r = requests.get(image_url)
    buffer_image = BytesIO(r.content)
    buffer_image.seek(0)
    res = make_response(send_file(buffer_image, mimetype='image/svg+xml'))
    res.headers['Cache-Control'] = \
        'max-age=0, no-cache, no-store, must-revalidate'
    res.headers['Expires'] = '0'
    return res


@blueprint.route("badge/<path:pipeline_id>")
def badge(pipeline_id):
    '''An individual pipeline status'''
    if not pipeline_id.startswith('./'):
        pipeline_id = './' + pipeline_id
    pipeline_status = status.get(pipeline_id)

    status_color = 'lightgray'
    if pipeline_status.pipeline_details:
        status_text = pipeline_status.state().lower()
        last_execution = pipeline_status.get_last_execution()
        success = last_execution.success if last_execution else None
        if success is True:
            stats = last_execution.stats if last_execution else None
            record_count = stats.get('count_of_rows')
            if record_count is not None:
                status_text += ' (%d records)' % record_count
            status_color = 'brightgreen'
        elif success is False:
            status_color = 'red'
    else:
        status_text = "not found"
    return _make_badge_response('pipeline', status_text, status_color)


@blueprint.route("badge/collection/<path:pipeline_path>")
def badge_collection(pipeline_path):
    '''Status badge for a collection of pipelines.'''
    all_pipeline_ids = sorted(status.all_pipeline_ids())

    if not pipeline_path.startswith('./'):
        pipeline_path = './' + pipeline_path

    # Filter pipeline ids to only include those that start with pipeline_path.
    path_pipeline_ids = \
        [p for p in all_pipeline_ids if p.startswith(pipeline_path)]

    statuses = []
    for pipeline_id in path_pipeline_ids:
        pipeline_status = status.get(pipeline_id)
        if pipeline_status is None:
            abort(404)
        status_text = pipeline_status.state().lower()
        statuses.append(status_text)

    status_color = 'lightgray'
    status_counter = Counter(statuses)
    if status_counter:
        if len(status_counter) == 1 and status_counter['succeeded'] > 0:
            status_color = 'brightgreen'
        elif status_counter['failed'] > 0:
            status_color = 'red'
        elif status_counter['failed'] == 0:
            status_color = 'yellow'
        status_text = \
            ', '.join(['{} {}'.format(v, k)
                       for k, v in status_counter.items()])
    else:
        status_text = "not found"

    return _make_badge_response('pipelines', status_text, status_color)


app = Flask(__name__)
app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True

if os.environ.get('DPP_BASIC_AUTH_USERNAME', False) \
   and os.environ.get('DPP_BASIC_AUTH_PASSWORD', False):
    app.config['BASIC_AUTH_USERNAME'] = os.environ['DPP_BASIC_AUTH_USERNAME']
    app.config['BASIC_AUTH_PASSWORD'] = os.environ['DPP_BASIC_AUTH_PASSWORD']
    app.config['BASIC_AUTH_ACTIVE'] = True

basic_auth = BasicAuth(app)


CORS(app)

url_prefix = os.environ.get('DPP_BASE_PATH', '/')
if not url_prefix.endswith('/'):
    url_prefix += '/'
logging.info('Serving on path %s', url_prefix)

app.register_blueprint(blueprint, url_prefix=url_prefix)



class BaseRunner(object):

    def __init__(self, name, parameters):
        self.name = name
        self.parameters = parameters

    def get_execution_args(self, step, cwd, idx):
        raise NotImplementedError()



import datetime
import json as _json

import decimal
import isodate

from .lazy_dict import LazyDict


DATE_FORMAT = '%Y-%m-%d'
DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
TIME_FORMAT = '%H:%M:%S'


class LazyJsonLine(LazyDict):

    def __init__(self, args, kwargs):
        super().__init__()
        self.line = args[0]
        self.args = args
        self.kwargs = kwargs

    def _evaluate(self):
        return json.loads(*self.args, **self.kwargs)

    def __str__(self):
        if self.inner is not None:
            return str(self.inner)
        return self.line

    def __repr__(self):
        if self.inner is not None:
            return repr(self.inner)
        return self.line


class CommonJSONDecoder(_json.JSONDecoder):
    """
    Common JSON Encoder
    json.loads(myString, cls=CommonJSONEncoder)
    """

    @classmethod
    def object_hook(cls, obj):
        if 'type{decimal}' in obj:
            try:
                return decimal.Decimal(obj['type{decimal}'])
            except decimal.InvalidOperation:
                pass
        if 'type{time}' in obj:
            try:
                return datetime.datetime \
                    .strptime(obj["type{time}"], TIME_FORMAT) \
                    .time()
            except ValueError:
                pass
        if 'type{datetime}' in obj:
            try:
                return datetime.datetime \
                    .strptime(obj["type{datetime}"], DATETIME_FORMAT)
            except ValueError:
                pass
        if 'type{date}' in obj:
            try:
                return datetime.datetime \
                    .strptime(obj["type{date}"], DATE_FORMAT) \
                    .date()
            except ValueError:
                pass
        if 'type{duration}' in obj:
            try:
                return isodate.parse_duration(obj["type{duration}"])
            except ValueError:
                pass
        if 'type{set}' in obj:
            try:
                return set(obj['type{set}'])
            except ValueError:
                pass

        return obj

    def __init__(self, **kwargs):
        kwargs['object_hook'] = self.object_hook
        super(CommonJSONDecoder, self).__init__(**kwargs)


class CommonJSONEncoder(_json.JSONEncoder):
    """
    Common JSON Encoder
    json.dumps(myString, cls=CommonJSONEncoder)
    """

    def default(self, obj):

        if isinstance(obj, decimal.Decimal):
            return {'type{decimal}': str(obj)}
        elif isinstance(obj, datetime.time):
            return {'type{time}': obj.strftime(TIME_FORMAT)}
        elif isinstance(obj, datetime.datetime):
            return {'type{datetime}': obj.strftime(DATETIME_FORMAT)}
        elif isinstance(obj, datetime.date):
            return {'type{date}': obj.strftime(DATE_FORMAT)}
        elif isinstance(obj, (isodate.Duration, datetime.timedelta)):
            return {'type{duration}': isodate.duration_isoformat(obj)}
        elif isinstance(obj, set):
            return {'type{set}': list(obj)}
        elif isinstance(obj, LazyDict):
            return obj.inner
        return super().default(obj)


def _dumpl(*args, **kwargs):
    obj = args[0]
    if isinstance(obj, LazyJsonLine):
        if not obj.dirty:
            return obj.line
        else:
            kwargs['cls'] = CommonJSONEncoder
            return _json.dumps(obj.inner, **kwargs)
    kwargs['cls'] = CommonJSONEncoder
    return _json.dumps(*args, **kwargs)


def _loadl(*args, **kwargs):
    if args[0][0] == '{':
        kwargs['cls'] = CommonJSONDecoder
        return LazyJsonLine(args, kwargs)
    else:
        return _loads(*args, **kwargs)


def _dumps(*args, **kwargs):
    kwargs['cls'] = CommonJSONEncoder
    return _json.dumps(*args, **kwargs)


def _loads(*args, **kwargs):
    kwargs['cls'] = CommonJSONDecoder
    return _json.loads(*args, **kwargs)


def _dump(*args, **kwargs):
    kwargs['cls'] = CommonJSONEncoder
    return _json.dump(*args, **kwargs)


def _load(*args, **kwargs):
    kwargs['cls'] = CommonJSONDecoder
    return _json.load(*args, **kwargs)


class json(object):
    dumps = _dumps
    loads = _loads
    dumpl = _dumpl
    loadl = _loadl
    dump = _dump
    load = _load
    JSONDecodeError = _json.JSONDecodeError



import os
import sys
import shlex

from ...utilities.extended_json import json
from .base_runner import BaseRunner


class LocalPythonRunner(BaseRunner):

    def get_execution_args(self, step, _, idx):
        return [
            sys.executable,
            step['executor'],
            str(idx),
            json.dumps(step.get('parameters', {})),
            str(step.get('validate', False)),
            step.get('_cache_hash') if step.get('cache') else ''
        ]


class WrappedPythonRunner(LocalPythonRunner):

    def get_execution_args(self, step, cwd, idx):
        args = super(WrappedPythonRunner, self).get_execution_args(step, cwd, idx)
        for i in range(len(args)):
            args[i] = '\\\"' + args[i].replace('"', '\\\\\\\"') + '\\\"'
        cmd = " ".join(args)
        abspath = os.path.abspath(cwd)
        cmd = self.parameters['wrapper'].format(path=cwd,
                                                abspath=abspath,
                                                cmd=cmd,
                                                env=os.environ)
        args = shlex.split(cmd)
        return args



import os
import sqlite3
from datapackage_pipelines.utilities.extended_json import json

DPP_DB_FILENAME = os.environ.get('DPP_DB_FILENAME', '.dpp.db')


class Sqlite3Dict(object):
    def __init__(self, filename):
        self.filename = filename
        conn = sqlite3.connect(self.filename)
        cursor = conn.cursor()
        cursor.execute(
            '''CREATE TABLE IF NOT EXISTS d (_key text, _value text)'''
        )
        conn.commit()
        conn.close()

    def __getitem__(self, key):
        conn = sqlite3.connect(self.filename)
        cursor = conn.cursor()
        result = cursor.execute(
            'SELECT _value from d where _key=?',
            (key,)
        ).fetchone()
        conn.close()
        if result is not None:
            return json.loads(result[0])
        return None

    def __setitem__(self, key, value):
        conn = sqlite3.connect(self.filename)
        value = json.dumps(value)
        cursor = conn.cursor()
        cursor.execute('DELETE FROM d where _key=?', (key,))
        cursor.execute('INSERT INTO d VALUES (?,?)', (key, value))
        conn.commit()
        conn.close()

    def __delitem__(self, key):
        conn = sqlite3.connect(self.filename)
        cursor = conn.cursor()
        cursor.execute('DELETE FROM d where _key=?', (key,))
        conn.commit()
        conn.close()


class SqliteBackend(object):

    KIND = 'sqlite3'
    ALL_PIPELINES_KEY = 'all-pipelines'

    def __init__(self):
        self.db = Sqlite3Dict(DPP_DB_FILENAME)

    def get_status(self, pipeline_id):
        return self.db[pipeline_id]

    def set_status(self, pipeline_id, status):
        self.db[pipeline_id] = status

    def del_status(self, pipeline_id):
        del self.db[pipeline_id]

    def register_pipeline_id(self, pipeline_id):
        all_pipelines = self.db[self.ALL_PIPELINES_KEY]
        if all_pipelines is None:
            all_pipelines = []
        if pipeline_id not in all_pipelines:
            all_pipelines.append(pipeline_id)
        self.db[self.ALL_PIPELINES_KEY] = all_pipelines

    def deregister_pipeline_id(self, pipeline_id):
        all_pipelines = self.db[self.ALL_PIPELINES_KEY]
        if all_pipelines is None:
            all_pipelines = []
        if pipeline_id in all_pipelines:
            all_pipelines = filter(lambda x: x != pipeline_id, all_pipelines)
        self.db[self.ALL_PIPELINES_KEY] = all_pipelines

    def reset(self):
        self.db[self.ALL_PIPELINES_KEY] = []

    def all_pipeline_ids(self):
        all_ids = sorted(self.db[self.ALL_PIPELINES_KEY])
        return all_ids

    def all_statuses(self):
        all_ids = sorted(self.db[self.ALL_PIPELINES_KEY])
        return [self.db[_id] for _id in all_ids]



import gzip
import sys
import os
import logging
from contextlib import ExitStack, redirect_stderr, redirect_stdout

from tableschema.exceptions import CastError

from ..utilities.extended_json import json

from .input_processor import process_input

from ..utilities.resources import streaming


logging.basicConfig(level=logging.DEBUG,
                    format="%(levelname)-8s:%(message)s")


cache = ''
first = True
stdout = sys.stdout

dependency_datapackage_urls = {}


def get_dependency_datapackage_url(pipeline_id):
    return dependency_datapackage_urls.get(pipeline_id)


# ## LOW LEVEL INTERFACE

def _ingest(debug=False):
    global cache
    global first
    params = None
    validate = False
    if len(sys.argv) > 4:
        first = sys.argv[1] == '0'
        params = json.loads(sys.argv[2])
        validate = sys.argv[3] == 'True'
        cache = sys.argv[4]

    datapackage, resource_iterator, dependency_dp = process_input(sys.stdin, validate, debug)
    dependency_datapackage_urls.update(dependency_dp)

    return params, datapackage, resource_iterator


def spew(dp, resources_iterator, stats=None, finalizer=None):
    files = [stdout]

    cache_filename = ''
    if len(cache) > 0:
        if not os.path.exists('.cache'):
            os.mkdir('.cache')
        cache_filename = os.path.join('.cache', cache)
        files.append(gzip.open(cache_filename+'.ongoing', 'wt'))

    expected_resources = \
        len(list(filter(streaming, dp.get('resources', []))))
    row_count = 0
    try:
        for f in files:
            f.write(json.dumps(dp, sort_keys=True, ensure_ascii=True)+'\n')
            f.flush()
        num_resources = 0
        for res in resources_iterator:
            if hasattr(res, 'it') and res.it is None:
                continue
            num_resources += 1
            for f in files:
                f.write('\n')
            try:
                for rec in res:
                    try:
                        line = json.dumpl(rec,
                                        sort_keys=True,
                                        ensure_ascii=True)
                    except TypeError as e:
                        logging.error('Failed to encode row to JSON: %s\nOffending row: %r', e, rec)
                        raise
                    # logging.error('SPEWING: {}'.format(line))
                    for f in files:
                        f.write(line+'\n')
                    # logging.error('WROTE')
                    row_count += 1
            except CastError as e:
                for err in e.errors:
                    logging.error('Failed to cast row: %s', err)
                raise
        if num_resources != expected_resources:
            logging.error('Expected to see %d resource(s) but spewed %d',
                          expected_resources, num_resources)
            assert num_resources == expected_resources

        aggregated_stats = {}
        if not first:
            stats_line = sys.stdin.readline().strip()
            if len(stats_line) > 0:
                try:
                    aggregated_stats = json.loads(stats_line)
                    assert aggregated_stats is None or \
                        isinstance(aggregated_stats, dict)
                except json.JSONDecodeError:
                    logging.error('Failed to parse stats: %r', stats_line)
        if stats is not None:
            aggregated_stats.update(stats)
        stats_json = json.dumps(aggregated_stats,
                                sort_keys=True,
                                ensure_ascii=True)
        for f in files:
            f.write('\n'+stats_json)

    except BrokenPipeError:
        logging.error('Output pipe disappeared!')
        sys.stderr.flush()
        sys.exit(1)

    stdout.flush()
    if row_count > 0:
        logging.info('Processed %d rows', row_count)

    if finalizer is not None:
        finalizer()

    for f in files:
        f.write('\n')  # Signal to other processors that we're done
        if f == stdout:
            # Can't close sys.stdout, otherwise any subsequent
            # call to print() will throw an exception
            f.flush()
        else:
            f.close()

    if len(cache) > 0:
        os.rename(cache_filename+'.ongoing', cache_filename)


class StdoutWriter:

    def write(self, message):
        message = message.strip()
        if message:
            logging.info(message)

    def flush(self):
        pass


class StderrWriter:

    def write(self, message):
        message = message.strip()
        if message:
            logging.error(message)

    def flush(self):
        pass


class ProcessorContext(ExitStack):

    def __init__(self, parameters, datapackage, resource_iterator):
        super().__init__()
        self.parameters = parameters
        self.datapackage = datapackage
        self.resource_iterator = resource_iterator
        self.stats = {}

    def __iter__(self):
        return iter((self.parameters, self.datapackage, self.resource_iterator))

    def __enter__(self):
        super().__enter__()
        self.enter_context(redirect_stdout(StdoutWriter()))
        self.enter_context(redirect_stderr(StderrWriter()))
        return self

    def __exit__(self, *args, **kw):
        spew(self.datapackage, self.resource_iterator, stats=self.stats)
        super().__exit__(*args, **kw)


def ingest(debug=False):
    params, datapackage, resource_iterator = _ingest(debug=debug)
    return ProcessorContext(params, datapackage, resource_iterator)


# ## HIGH LEVEL INTERFACE

def generic_process_resource(rows,
                             spec,
                             resource_index,
                             parameters,
                             stats,
                             process_row):
    for row_index, row in enumerate(rows):
        row = process_row(row, row_index,
                          spec, resource_index,
                          parameters, stats)
        if row is not None:
            yield row


def generic_process_resources(resource_iterator,
                              parameters,
                              stats,
                              process_row):
    for resource_index, resource in enumerate(resource_iterator):
        rows = resource
        spec = resource.spec
        yield generic_process_resource(rows,
                                       spec,
                                       resource_index,
                                       parameters,
                                       stats,
                                       process_row)


def process(modify_datapackage=None,
            process_row=None, debug=False):
    with ingest(debug=debug) as ctx:
        if modify_datapackage is not None:
            ctx.datapackage = modify_datapackage(ctx.datapackage, ctx.parameters, ctx.stats)

        if process_row is not None:
            ctx.resource_iterator = \
                generic_process_resources(ctx.resource_iterator,
                                          ctx.parameters,
                                          ctx.stats,
                                          process_row)
